#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include "MagneticDisk.h"


/*--------------------- VITERBI DETECTOR FOR 6/7 MTR CODE -------------------------------*/
void Viterbi_6By7MTRCode(int Delay, double *EqualizedChannelOutput, int PrePaddingLength, int SectorLength, int *DetectedOutput, double *PRImpRes, int PRImpResLength)
{
	int i,j,k,m,NumStates,NumCodeword;
	double metric0,metric1;
	double sample0,sample1;
	double *PathMetric[2];
	int **Path[2];
	int *VS[2];
	int MinPathMetricIndex;
	int VSLengthPrevTime=1,VSLengthCurrentTime=0;

	/* Number of bits in every state = PRImpResLength-1 (= Order of PR polynomial) */
	/* Number of states in trellis = 2^(Order of PR polynomial) */
	NumStates = (int)pow(2,(PRImpResLength-1));

	/* 
	Path[0][i][j] : Bit at a delay (Delay-j) of the survivor path corresponding to state i
					at the previous time.
	Path[1][i][j] : Bit at a delay (Delay-j) of the survivor path corresponding to state i
					at the current time.
	*/
	Path[0] = (int**)malloc(NumStates*sizeof(int));
	Path[1] = (int**)malloc(NumStates*sizeof(int));
	for(i=0;i<NumStates;i++)
	{
		Path[0][i] = (int*)malloc(Delay*sizeof(int));
		Path[1][i] = (int*)malloc(Delay*sizeof(int));
	}
	for(i=0;i<NumStates;i++)
	{
		for(j=0;j<Delay;j++)
		{
			Path[0][i][j] = 0;
			Path[1][i][j] = 0;
		}
	}
	PathMetric[0] = (double *)malloc(NumStates * sizeof(double));
	PathMetric[1] = (double *)malloc(NumStates * sizeof(double));
	/* PathMetric[i] => The path metric of the valid output of state i at current time */
	/*
	NOTE that Pathmetric is initialized to 0 in the hope that by the time 'Delay' bits are
	transmitted, all of these are overwritten. If not, then when making a decision on a bit,
	the state for which Pathmetric is still 0 will be chosen, which is wrong. It cannot be
	initialized to a large number to avoid this problem because it is additive.
	*/
	for(i=0;i<NumStates;i++)
	{
		PathMetric[0][i] = 0.0;
		PathMetric[1][i] = 0.0;
	}
	
	/* Initially, until time (PRImpResLength-1), all states are not valid */
	/* VS[0] => Valid States in the previous time */
	/* VS[1] => Valid States in the current time */
	VS[0] = (int *)malloc(sizeof(int));
	VS[1] = (int *)malloc(sizeof(int));
	VS[0][0]=0;	/* Assuming that we start with state 0 at time 0 */
	VS[1][0]=0;

			/** Adapt Viterbi according to MTR constraints **/
			/* Check if the current state is a valid state for the 6/7 MTR(2;8) code */
			if(PRImpResLength-1 == 2)
			{
				printf("\nERROR: -> Viterbi() : Viterbi not defined for PR Impulse response length = 3 \n");
				exit(-1);

			}else if(PRImpResLength-1 == 3)
			{
				/* All states are valid */
			}else if(PRImpResLength-1 == 4)
			{
				PathMetric[0][5] = 1e50;	/* Some large number, so that this path will never be selected */
				PathMetric[0][10] = 1e50;	/* Some large number, so that this path will never be selected */

			}else if(PRImpResLength-1 == 5)
			{
				printf("\nERROR: -> Viterbi() : Viterbi not defined for PR impulse response length > 6 \n");
				exit(-1);
				//if(i==5 || i==10 || i==11 || i==20 || i==21 || i==36)
				//	PathMetric[1][i] = 1e50;
			}
			else
			{
				printf("\nERROR: -> Viterbi() : Viterbi not defined for PR impulse response length > 6 \n");
				exit(-1);
			}

	/**** Main VITERBI loop ****/
	for(k=0;k<SectorLength;k++)
	{/* Loop over PaddedEqOutput */

		for(i=0;i<NumStates;i++)
		{/* Loop over all states */

			/*** For each state, calculating the path metric of the two incoming paths ***/

			/*
			If i = [i1 i2 ... in] (the array elements are binary) is the current state
			then its corresponding previous states are [0 i1 i2.....in-1] and [1 i1 i2....in-1].
			These are determined by	bit operations given below.
			*/

			/**
			Calculating the sample value and metric at current time for the transition from
			state [0 i1 i2....in-1]	to state [i1 i2....in].
			Corresponding Previous state is given by: (CurrentState >> 1) => shifting all
			bits to the right by 1 position and appending a ZERO at the left most position.
			**/
			sample0 = 0.0;
			for(j=0;j<PRImpResLength-1;j++)
			{
				sample0 += PRImpRes[j]*((i >> j) & 1);
			}
			sample0 += PRImpRes[PRImpResLength-1]*0;	/* Eventhough it is 0, it is explicitly written for clarity */
			sample0 = sample0*2;
			metric0 = PathMetric[0][i >> 1] + pow((EqualizedChannelOutput[k] - sample0),2);

			/**
			Calculating the sample value and metric at current time for the transition from
			state [1 i1 i2....in-1]	to state [i1 i2....in]
			Corresponding Previous state is given by: (CurrentState >> 1) | (1 << (CurrentStateLength-1))
			=> shifting all bits to the right by 1 position and appending a ONE at the
			left most position
			**/
			sample1 = 0.0;
			for(j=0;j<PRImpResLength-1;j++)
			{
				sample1 += PRImpRes[j]*((i >> j) & 1);
			}
			sample1 += PRImpRes[PRImpResLength-1]*1;
			sample1 = sample1*2;
			metric1 = PathMetric[0][(i >> 1)|(1 << (PRImpResLength-2))] + pow((EqualizedChannelOutput[k] - sample1),2);

			/* 
			Assuming that we start with state 0, then until time (PRImpResLength-1),
			all states will have less than 2 valid incoming paths. For example, at time 1, the 
			incoming paths for any state is valid only if it comes from state 0, because we 
			assume that we start from state 0 in time 0.
			*/
			if(k < (PRImpResLength-1))
			{
				for(j=0;j<VSLengthPrevTime;j++)
				{
					if((i >> 1)==VS[0][j])
					{	
						/* This is the only valid incoming path */
						for(m=0;m<Delay-1;m++)
						{
							Path[1][i][m] = Path[0][i >> 1][m+1];
						}
						Path[1][i][Delay-1] = i & 1;
						//Path[1][i] = (Path[0][i >> 1] << 1) | (i & 1);
						PathMetric[1][i] = metric0;
						VSLengthCurrentTime++;
						VS[1] = (int*)realloc(VS[1],(VSLengthCurrentTime)*sizeof(int));
						/* 
						This state now becomes valid, i.e. any path emerging out of
						this state is valid in the next time
						*/
						VS[1][VSLengthCurrentTime-1] = i;
						break;	/* At most only one path is valid */
					}
					if(((i >> 1) | (1 << (PRImpResLength-2))) == VS[0][j])
					{
						/* This is the only valid incoming path */
						for(m=0;m<Delay-1;m++)
						{
							Path[1][i][m] = Path[0][(i >> 1)|(1 << (PRImpResLength-2))][m+1];
						}
						Path[1][i][Delay-1] = i & 1;						
						//Path[1][i] = (Path[0][(i >> 1)|(1 << (PRImpResLength-2))] << 1) | (i & 1);
						PathMetric[1][i] = metric1;
						VSLengthCurrentTime++;
						VS[1] = (int*)realloc(VS[1],(VSLengthCurrentTime)*sizeof(int));
						/* 
						This state now becomes valid, i.e. any path emerging out of
						this state is valid in the next time
						*/
						VS[1][VSLengthCurrentTime-1] = i;
						break;	/* At most only one path is valid */
					}
				}
			}
			else
			/* All incoming paths for all states are valid */
			{
				if(metric0 <= metric1)
				{
					/*
					New path is from the state [0 i1 i2...in-1], i.e current valid path for state i is
					the valid path of state [0 i1...in-1] in the previous time appended with bit 'in'
					*/
					/*
					(Path[0][i >> 1] << 1) => shifting all bits of path of the state [0 i1...in-1]
					to the left by 1 position.
					(i & 1) => The last bit of state i.
					(Path[0][i >> 1] << 1) | (i & 1) => appending last bit of i to the left-shifted path.
					*/
					for(m=0;m<Delay-1;m++)
					{
						Path[1][i][m] = Path[0][i >> 1][m+1];
					}
					Path[1][i][Delay-1] = i & 1;
					//Path[1][i] = (Path[0][i >> 1] << 1) | (i & 1);
					PathMetric[1][i] = metric0;
				}else{
					/*
					New path is from the state [1 i1 i2...in-1], i.e current valid path for state i is
					the valid path of state [1 i1...in-1] in the previous time appended with bit 'in'
					*/
					/* Bit operations are done as before */
					for(m=0;m<Delay-1;m++)
					{
						Path[1][i][m] = Path[0][(i >> 1)|(1 << (PRImpResLength-2))][m+1];
					}
					Path[1][i][Delay-1] = i & 1;
					//Path[1][i] = (Path[0][(i >> 1)|(1 << (PRImpResLength-2))] << 1) | (i & 1);
					PathMetric[1][i] = metric1;
				}
			}

			if((k>=PrePaddingLength+1) && (k<(PrePaddingLength+SectorLength)))
			{
				/* Count the number of codewords */
				NumCodeword = (k-PrePaddingLength-1)/7 + 1;	/* Starting from 1 */

				/** Adapt Viterbi according to MTR constraints **/
				/* Check if the current state is a valid state for the 6/7 MTR(2;8) code */
				if(PRImpResLength-1 == 2)
				{
					printf("\nERROR: -> Viterbi() : Viterbi not defined for PR Impulse response length = 3 \n");
					exit(-1);

				}else if(PRImpResLength-1 == 3)
				{
					/* All states are valid */
				}else if(PRImpResLength-1 == 4)
				{
					if(i==5 || i==10)				/* Invalid states: 0101 & 1010 */
						PathMetric[1][i] = 1e50;	/* Some large number, so that this path will never be selected */
					//if(((k-PrePaddingLength)-(NumCodeword-1)*7) == 7 || ((k-PrePaddingLength)-(NumCodeword-1)*7) == 2)
					//{
					//	if(i==2 || i==13)
					//		PathMetric[1][i] = 1e50;	/* Some large number, so that this path will never be selected */
					//}

				}else if(PRImpResLength-1 == 5)
				{
					printf("\nERROR: -> Viterbi() : Viterbi not defined for PR impulse response length > 6 \n");
					exit(-1);
					//if(i==5 || i==10 || i==11 || i==20 || i==21 || i==36)
					//	PathMetric[1][i] = 1e50;
				}
				else
				{
					printf("\nERROR: -> Viterbi() : Viterbi not defined for PR impulse response length > 6 \n");
					exit(-1);
				}
			}
		}/* Loop over all states */

		/* Make a decision on the bit transmitted 'Delay' bit periods before */
		if(k >= Delay-1)
		{
			/** Deciding on the transmitted bit based on detection delay **/
			/* Finding the minimum path metric at the current time */
			MinPathMetricIndex = 0;
			for(i=1;i<NumStates;i++)
			{
				if(PathMetric[1][i] < PathMetric[1][MinPathMetricIndex])
					MinPathMetricIndex = i;
			}
			/* Detected output = first bit in Path[1][MinPathMetricIndex] */
			/* Path[1][MinPathMetricIndex] has 'Delay' number of bits */
			DetectedOutput[k-(Delay-1)] = Path[1][MinPathMetricIndex][0];
		}

		/* Exchange old and new Path and PathMetric values */
		for(i=0;i<NumStates;i++)
		{
			for(m=0;m<Delay;m++)
			{
				Path[0][i][m] = Path[1][i][m];
				Path[1][i][m] = 0;
			}
			PathMetric[0][i] = PathMetric[1][i];
			PathMetric[1][i] = 0.0;
		}
		
		/* Exchange old and new Valid State information */
		if(k < (PRImpResLength - 1))
		{
			VS[0] = (int*)realloc(VS[0],VSLengthCurrentTime*sizeof(int));
			for(i=0;i<VSLengthCurrentTime;i++)
			{
				VS[0][i] = VS[1][i];
				VS[1][i] = 0;
			}
			VSLengthPrevTime = VSLengthCurrentTime;
			VSLengthCurrentTime = 0;
		}

	}/* Loop over PaddedEqOutput */

	/* Detecting the last 'delay-1' bits */
	for(i=2;i<=Delay;i++)
	{
		DetectedOutput[SectorLength-1-(Delay-i)] = Path[0][MinPathMetricIndex][i-1];
	}
	for(i=0;i<NumStates;i++)
	{
		free(Path[0][i]);
		free(Path[1][i]);
	}
	free(Path[0]);
	free(Path[1]);
	free(VS[0]);
	free(VS[1]);
	free(PathMetric[0]);
	free(PathMetric[1]);
}



/*--------------------- VITERBI DETECTOR FOR 8/9 MTR CODE -------------------------------*/


void Viterbi_8By9TMTRCode(int Delay, double *EqualizedChannelOutput, int PrePaddingLength, int SectorLength, int *DetectedOutput, double *PRImpRes, int PRImpResLength)
{
	int i,j,k,m,NumStates,NumCodeword;
	double metric0,metric1;
	double sample0,sample1;
	double *PathMetric[2];
	int **Path[2];
	int *VS[2];
	int MinPathMetricIndex;
	int VSLengthPrevTime=1,VSLengthCurrentTime=0;

	/* Number of bits in every state = PRImpResLength-1 (= Order of PR polynomial) */
	/* Number of states in trellis = 2^(Order of PR polynomial) */
	NumStates = (int)pow(2,(PRImpResLength-1));

	/* 
	Path[0][i][j] : Bit at a delay (Delay-j) of the survivor path corresponding to state i
					at the previous time.
	Path[1][i][j] : Bit at a delay (Delay-j) of the survivor path corresponding to state i
					at the current time.
	*/
	Path[0] = (int**)malloc(NumStates*sizeof(int));
	Path[1] = (int**)malloc(NumStates*sizeof(int));
	for(i=0;i<NumStates;i++)
	{
		Path[0][i] = (int*)malloc(Delay*sizeof(int));
		Path[1][i] = (int*)malloc(Delay*sizeof(int));
	}
	for(i=0;i<NumStates;i++)
	{
		for(j=0;j<Delay;j++)
		{
			Path[0][i][j] = 0;
			Path[1][i][j] = 0;
		}
	}
	PathMetric[0] = (double *)malloc(NumStates * sizeof(double));
	PathMetric[1] = (double *)malloc(NumStates * sizeof(double));
	/* PathMetric[i] => The path metric of the valid output of state i at current time */
	/*
	NOTE that Pathmetric is initialized to 0 in the hope that by the time 'Delay' bits are
	transmitted, all of these are overwritten. If not, then when making a decision on a bit,
	the state for which Pathmetric is still 0 will be chosen, which is wrong. It cannot be
	initialized to a large number to avoid this problem because it is additive.
	*/
	for(i=0;i<NumStates;i++)
	{
		PathMetric[0][i] = 0.0;
		PathMetric[1][i] = 0.0;
	}
	
	/* Initially, until time (PRImpResLength-1), all states are not valid */
	/* VS[0] => Valid States in the previous time */
	/* VS[1] => Valid States in the current time */
	VS[0] = (int *)malloc(sizeof(int));
	VS[1] = (int *)malloc(sizeof(int));
	VS[0][0]=0;	/* Assuming that we start with state 0 at time 0 */
	VS[1][0]=0;

			/** Adapt Viterbi according to MTR constraints **/
			/* Check if the current state is a valid state for the 6/7 MTR(2;8) code */
			if(PRImpResLength-1 == 2)
			{
				printf("\nERROR: -> Viterbi() : Viterbi not defined for PR Impulse response length = 3 \n");
				exit(-1);

			}else if(PRImpResLength-1 == 3)
			{
				/* All states are valid */
			}else if(PRImpResLength-1 == 4)
			{
				PathMetric[0][5] = 1e50;	/* Some large number, so that this path will never be selected */
				PathMetric[0][10] = 1e50;	/* Some large number, so that this path will never be selected */

			}else if(PRImpResLength-1 == 5)
			{
				printf("\nERROR: -> Viterbi() : Viterbi not defined for PR impulse response length > 6 \n");
				exit(-1);
				//if(i==5 || i==10 || i==11 || i==20 || i==21 || i==36)
				//	PathMetric[1][i] = 1e50;
			}
			else
			{
				printf("\nERROR: -> Viterbi() : Viterbi not defined for PR impulse response length > 6 \n");
				exit(-1);
			}

	/**** Main VITERBI loop ****/
	for(k=0;k<SectorLength;k++)
	{/* Loop over PaddedEqOutput */

		for(i=0;i<NumStates;i++)
		{/* Loop over all states */

			/*** For each state, calculating the path metric of the two incoming paths ***/

			/*
			If i = [i1 i2 ... in] (the array elements are binary) is the current state
			then its corresponding previous states are [0 i1 i2.....in-1] and [1 i1 i2....in-1].
			These are determined by	bit operations given below.
			*/

			/**
			Calculating the sample value and metric at current time for the transition from
			state [0 i1 i2....in-1]	to state [i1 i2....in].
			Corresponding Previous state is given by: (CurrentState >> 1) => shifting all
			bits to the right by 1 position and appending a ZERO at the left most position.
			**/
			sample0 = 0.0;
			for(j=0;j<PRImpResLength-1;j++)
			{
				sample0 += PRImpRes[j]*((i >> j) & 1);
			}
			sample0 += PRImpRes[PRImpResLength-1]*0;	/* Eventhough it is 0, it is explicitly written for clarity */
			sample0 = sample0*2;
			metric0 = PathMetric[0][i >> 1] + pow((EqualizedChannelOutput[k] - sample0),2);

			/**
			Calculating the sample value and metric at current time for the transition from
			state [1 i1 i2....in-1]	to state [i1 i2....in]
			Corresponding Previous state is given by: (CurrentState >> 1) | (1 << (CurrentStateLength-1))
			=> shifting all bits to the right by 1 position and appending a ONE at the
			left most position
			**/
			sample1 = 0.0;
			for(j=0;j<PRImpResLength-1;j++)
			{
				sample1 += PRImpRes[j]*((i >> j) & 1);
			}
			sample1 += PRImpRes[PRImpResLength-1]*1;
			sample1 = sample1*2;
			metric1 = PathMetric[0][(i >> 1)|(1 << (PRImpResLength-2))] + pow((EqualizedChannelOutput[k] - sample1),2);

			/* 
			Assuming that we start with state 0, then until time (PRImpResLength-1),
			all states will have less than 2 valid incoming paths. For example, at time 1, the 
			incoming paths for any state is valid only if it comes from state 0, because we 
			assume that we start from state 0 in time 0.
			*/
			if(k < (PRImpResLength-1))
			{
				for(j=0;j<VSLengthPrevTime;j++)
				{
					if((i >> 1)==VS[0][j])
					{	
						/* This is the only valid incoming path */
						for(m=0;m<Delay-1;m++)
						{
							Path[1][i][m] = Path[0][i >> 1][m+1];
						}
						Path[1][i][Delay-1] = i & 1;
						//Path[1][i] = (Path[0][i >> 1] << 1) | (i & 1);
						PathMetric[1][i] = metric0;
						VSLengthCurrentTime++;
						VS[1] = (int*)realloc(VS[1],(VSLengthCurrentTime)*sizeof(int));
						/* 
						This state now becomes valid, i.e. any path emerging out of
						this state is valid in the next time
						*/
						VS[1][VSLengthCurrentTime-1] = i;
						break;	/* At most only one path is valid */
					}
					if(((i >> 1) | (1 << (PRImpResLength-2))) == VS[0][j])
					{
						/* This is the only valid incoming path */
						for(m=0;m<Delay-1;m++)
						{
							Path[1][i][m] = Path[0][(i >> 1)|(1 << (PRImpResLength-2))][m+1];
						}
						Path[1][i][Delay-1] = i & 1;						
						//Path[1][i] = (Path[0][(i >> 1)|(1 << (PRImpResLength-2))] << 1) | (i & 1);
						PathMetric[1][i] = metric1;
						VSLengthCurrentTime++;
						VS[1] = (int*)realloc(VS[1],(VSLengthCurrentTime)*sizeof(int));
						/* 
						This state now becomes valid, i.e. any path emerging out of
						this state is valid in the next time
						*/
						VS[1][VSLengthCurrentTime-1] = i;
						break;	/* At most only one path is valid */
					}
				}
			}
			else
			/* All incoming paths for all states are valid */
			{
				if(metric0 <= metric1)
				{
					/*
					New path is from the state [0 i1 i2...in-1], i.e current valid path for state i is
					the valid path of state [0 i1...in-1] in the previous time appended with bit 'in'
					*/
					/*
					(Path[0][i >> 1] << 1) => shifting all bits of path of the state [0 i1...in-1]
					to the left by 1 position.
					(i & 1) => The last bit of state i.
					(Path[0][i >> 1] << 1) | (i & 1) => appending last bit of i to the left-shifted path.
					*/
					for(m=0;m<Delay-1;m++)
					{
						Path[1][i][m] = Path[0][i >> 1][m+1];
					}
					Path[1][i][Delay-1] = i & 1;
					//Path[1][i] = (Path[0][i >> 1] << 1) | (i & 1);
					PathMetric[1][i] = metric0;
				}else{
					/*
					New path is from the state [1 i1 i2...in-1], i.e current valid path for state i is
					the valid path of state [1 i1...in-1] in the previous time appended with bit 'in'
					*/
					/* Bit operations are done as before */
					for(m=0;m<Delay-1;m++)
					{
						Path[1][i][m] = Path[0][(i >> 1)|(1 << (PRImpResLength-2))][m+1];
					}
					Path[1][i][Delay-1] = i & 1;
					//Path[1][i] = (Path[0][(i >> 1)|(1 << (PRImpResLength-2))] << 1) | (i & 1);
					PathMetric[1][i] = metric1;
				}
			}

			/* 
			Check for state validity for the data sequence.
			Note that we also ignore the first bit of the data sequence apart from
			the padded bits. This is because of the NRZI-NRZ conversion.
			*/	
			if((k>=PrePaddingLength+1) && (k<(PrePaddingLength+SectorLength)))
			{
				/* Count the number of codewords */
				NumCodeword = (k-PrePaddingLength-1)/9 + 1;	/* Starting from 1 */

				/* Check if the current state is a valid state for the 8/9 MTR code */
				if(PRImpResLength-1 == 2)
				{
					printf("\nERROR: -> Viterbi() : Viterbi not defined for PR Impulse response length = 3 \n");
					exit(-1);

				}else if(PRImpResLength-1 == 3)
				{
					/* All states are valid. But certain edges aren't, which is taken care elsewhere. */
				}else if(PRImpResLength-1 == 4)
				{
					/* For 8/9 MTR code, tribits are allowed only at odd bit locations. Since, the codeword length is 
					odd(9), valid states in a viterbi detector will change from codeword to codeword, apart from changing
					at every alternate times. That is, for the 1st codeword, tribits at odd time units will be allowed 
					but for the 2nd codeword tribits at even time units will be allowed.  Now, since 
					viterbi operates on NRZ bits, whose length is one more than the NRZI bits, allowed tribit locations
					are switched(even <-> odd). An another issue to worry about is the length of the prepadding bits. 
					If it is even, then as mentioned above, for NRZ bits, tribits locations are switched. 
					If it is odd, then for NRZ bits, tribit locations remain the same as for NRZI. Here, it is assumed that
					prepadding bit length is even. To summarise (assuming even prepadding length):

					NRZI: 
						@ Odd Codewords(starting from 1): Tribits at EVEN locations allowed
						@ Even Codewords(starting from 1): Tribits at ODD locations allowed

					NRZ: 
						@ Odd Codewords(starting from 1): Tribits at ODD locations allowed
						@ Even Codewords(starting from 1): Tribits at EVEN locations allowed
					*/

					/* Detecting Odd NRZI bits (or Even NRZ bits) */
					if(((k-PrePaddingLength)-(NumCodeword-1)*9)%2 == 1)// || ((k-PrePaddingLength)-(NumCodeword-1)*9) == 2)
					{
						if(i==5 || i==10)				/* Invalid states: 0101 & 1010 */
							PathMetric[1][i] = 1e50;	/* Some large number, so that this path will never be selected */
						/* Detecting end of codewords */
						/* All states that end in NRZI 11 are invalid */
						//if(((k-PrePaddingLength)-(NumCodeword-1)*9) == 9)
						//{
						//	if(i==2 || i==13)
						//		PathMetric[1][i] = 1e50;	/* Some large number, so that this path will never be selected */
						//}
					}else
					{
						/* All states are valid */

					}
				}else if(PRImpResLength-1 == 5)
				{

					printf("\nERROR: -> Viterbi() : Viterbi not defined for PR impulse response length = 6 \n");
					exit(-1);
				}
				else
				{
					printf("\nERROR: -> Viterbi() : Viterbi not defined for PR impulse response length > 6 \n");
					exit(-1);
				}
			}

		}/* Loop over all states */

		/* Make a decision on the bit transmitted 'Delay' bit periods before */
		if(k >= Delay-1)
		{
			/** Deciding on the transmitted bit based on detection delay **/
			/* Finding the minimum path metric at the current time */
			MinPathMetricIndex = 0;
			for(i=1;i<NumStates;i++)
			{
				if(PathMetric[1][i] < PathMetric[1][MinPathMetricIndex])
					MinPathMetricIndex = i;
			}
			/* Detected output = first bit in Path[1][MinPathMetricIndex] */
			/* Path[1][MinPathMetricIndex] has 'Delay' number of bits */
			DetectedOutput[k-(Delay-1)] = Path[1][MinPathMetricIndex][0];
		}

		/* Exchange old and new Path and PathMetric values */
		for(i=0;i<NumStates;i++)
		{
			for(m=0;m<Delay;m++)
			{
				Path[0][i][m] = Path[1][i][m];
				Path[1][i][m] = 0;
			}
			PathMetric[0][i] = PathMetric[1][i];
			PathMetric[1][i] = 0.0;
		}
		
		/* Exchange old and new Valid State information */
		if(k < (PRImpResLength - 1))
		{
			VS[0] = (int*)realloc(VS[0],VSLengthCurrentTime*sizeof(int));
			for(i=0;i<VSLengthCurrentTime;i++)
			{
				VS[0][i] = VS[1][i];
				VS[1][i] = 0;
			}
			VSLengthPrevTime = VSLengthCurrentTime;
			VSLengthCurrentTime = 0;
		}

	}/* Loop over PaddedEqOutput */

	/* Detecting the last 'delay-1' bits */
	for(i=2;i<=Delay;i++)
	{
		DetectedOutput[SectorLength-1-(Delay-i)] = Path[0][MinPathMetricIndex][i-1];
	}
	for(i=0;i<NumStates;i++)
	{
		free(Path[0][i]);
		free(Path[1][i]);
	}
	free(Path[0]);
	free(Path[1]);
	free(VS[0]);
	free(VS[1]);
	free(PathMetric[0]);
	free(PathMetric[1]);
}


/*---------------------------- VITERBI DETECTOR (DIFF IMPLEMENTATION) ------------------------*/

/* IMPORTANT: Use this version of viterbi if the Viterbi Delay is more than 20 */

void ClassicalViterbi(int Delay, double *EqualizedChannelOutput, int SectorLength, int *DetectedOutput, double *PRImpRes, int PRImpResLength)
{
	int i,j,k,m,NumStates;
	double metric0,metric1;
	double sample0,sample1;
	double *PathMetric[2];
	int **Path[2];
	int MinPathMetricIndex;

	/* Number of bits in every state = PRImpResLength-1 (= Order of PR polynomial) */
	/* Number of states in trellis = 2^(Order of PR polynomial) */
	NumStates = (int)pow(2,(PRImpResLength-1));

	/* 
	Path[0][i][j] : Bit at a delay (Delay-j) of the survivor path corresponding to state i
					at the previous time.
	Path[1][i][j] : Bit at a delay (Delay-j) of the survivor path corresponding to state i
					at the current time.
	*/
	Path[0] = (int**)malloc(NumStates*sizeof(int));
	Path[1] = (int**)malloc(NumStates*sizeof(int));
	for(i=0;i<NumStates;i++)
	{
		Path[0][i] = (int*)malloc(Delay*sizeof(int));
		Path[1][i] = (int*)malloc(Delay*sizeof(int));
	}
	for(i=0;i<NumStates;i++)
	{
		for(j=0;j<Delay;j++)
		{
			Path[0][i][j] = 0;
			Path[1][i][j] = 0;
		}
	}
	PathMetric[0] = (double *)malloc(NumStates * sizeof(double));
	PathMetric[1] = (double *)malloc(NumStates * sizeof(double));
	/* PathMetric[i] => The path metric of the valid output of state i at current time */
	/*
	NOTE that Pathmetric is initialized to 0 in the hope that by the time 'Delay' bits are
	transmitted, all of these are overwritten. If not, then when making a decision on a bit,
	the state for which Pathmetric is still 0 will be chosen, which is wrong. It cannot be
	initialized to a large number to avoid this problem because it is additive.
	*/
	for(i=0;i<NumStates;i++)
	{
		PathMetric[0][i] = 1e50;
		PathMetric[1][i] = 1e50;
	}
	PathMetric[0][0] = 0;
	PathMetric[1][0] = 0;
	
	/**** Main VITERBI loop ****/
	for(k=0;k<SectorLength;k++)
	{/* Loop over PaddedEqOutput */

		for(i=0;i<NumStates;i++)
		{/* Loop over all states */

			/*** For each state, calculating the path metric of the two incoming paths ***/

			/*
			If i = [i1 i2 ... in] (the array elements are binary) is the current state
			then its corresponding previous states are [0 i1 i2.....in-1] and [1 i1 i2....in-1].
			These are determined by	bit operations given below.
			*/

			/**
			Calculating the sample value and metric at current time for the transition from
			state [0 i1 i2....in-1]	to state [i1 i2....in].
			Corresponding Previous state is given by: (CurrentState >> 1) => shifting all
			bits to the right by 1 position and appending a ZERO at the left most position.
			**/
			sample0 = 0.0;
			for(j=0;j<PRImpResLength-1;j++)
			{
				sample0 += PRImpRes[j]*((i >> j) & 1);
			}
			sample0 += PRImpRes[PRImpResLength-1]*0;	/* Eventhough it is 0, it is explicitly written for clarity */
			sample0 = sample0*2;
			metric0 = PathMetric[0][i >> 1] + pow((EqualizedChannelOutput[k] - sample0),2);

			/**
			Calculating the sample value and metric at current time for the transition from
			state [1 i1 i2....in-1]	to state [i1 i2....in]
			Corresponding Previous state is given by: (CurrentState >> 1) | (1 << (CurrentStateLength-1))
			=> shifting all bits to the right by 1 position and appending a ONE at the
			left most position
			**/
			sample1 = 0.0;
			for(j=0;j<PRImpResLength-1;j++)
			{
				sample1 += PRImpRes[j]*((i >> j) & 1);
			}
			sample1 += PRImpRes[PRImpResLength-1]*1;
			sample1 = sample1*2;
			metric1 = PathMetric[0][(i >> 1)|(1 << (PRImpResLength-2))] + pow((EqualizedChannelOutput[k] - sample1),2);

			if(metric0 <= metric1)
			{
				/*
				New path is from the state [0 i1 i2...in-1], i.e current valid path for state i is
				the valid path of state [0 i1...in-1] in the previous time appended with bit 'in'
				*/
				/*
				(Path[0][i >> 1] << 1) => shifting all bits of path of the state [0 i1...in-1]
				to the left by 1 position.
				(i & 1) => The last bit of state i.
				(Path[0][i >> 1] << 1) | (i & 1) => appending last bit of i to the left-shifted path.
				*/
				for(m=0;m<Delay-1;m++)
				{
					Path[1][i][m] = Path[0][i >> 1][m+1];
				}
				Path[1][i][Delay-1] = i & 1;
				PathMetric[1][i] = metric0;
			}else{
				/*
				New path is from the state [1 i1 i2...in-1], i.e current valid path for state i is
				the valid path of state [1 i1...in-1] in the previous time appended with bit 'in'
				*/
				/* Bit operations are done as before */
				for(m=0;m<Delay-1;m++)
				{
					Path[1][i][m] = Path[0][(i >> 1)|(1 << (PRImpResLength-2))][m+1];
				}
				Path[1][i][Delay-1] = i & 1;
				PathMetric[1][i] = metric1;
			}
		}/* Loop over all states */

		/* Make a decision on the bit transmitted 'Delay' bit periods before */
		if(k >= Delay-1)
		{
			/** Deciding on the transmitted bit based on detection delay **/
			/* Finding the minimum path metric at the current time */
			MinPathMetricIndex = 0;
			for(i=1;i<NumStates;i++)
			{
				if(PathMetric[1][i] < PathMetric[1][MinPathMetricIndex])
					MinPathMetricIndex = i;
			}
			/* Detected output = first bit in Path[1][MinPathMetricIndex] */
			/* Path[1][MinPathMetricIndex] has 'Delay' number of bits */
			DetectedOutput[k-(Delay-1)] = Path[1][MinPathMetricIndex][0];
		}

		/* Exchange old and new Path and PathMetric values */
		for(i=0;i<NumStates;i++)
		{
			for(m=0;m<Delay;m++)
			{
				Path[0][i][m] = Path[1][i][m];
				Path[1][i][m] = 0;
			}
			PathMetric[0][i] = PathMetric[1][i];
			PathMetric[1][i] = 0.0;
		}
	}/* Loop over PaddedEqOutput */

	/* Detecting the last 'delay-1' bits */
	for(i=2;i<=Delay;i++)
	{
		DetectedOutput[SectorLength-1-(Delay-i)] = Path[0][MinPathMetricIndex][i-1];
	}
	for(i=0;i<NumStates;i++)
	{
		free(Path[0][i]);
		free(Path[1][i]);
	}
	free(Path[0]);
	free(Path[1]);
	free(PathMetric[0]);
	free(PathMetric[1]);
}


/*---------------------------- VITERBI DETECTOR (DIFF IMPLEMENTATION) ------------------------*/

/* IMPORTANT: Use this version of viterbi if the Viterbi Delay is more than 20 */

void ClassicalViterbi_4By5RLLCode(int Delay, double *EqualizedChannelOutput, int SectorLength, int *DetectedOutput, double *PRImpRes, int PRImpResLength)
{
	int i,j,k,m,NumStates;
	double metric0,metric1;
	double sample0,sample1;
	double *PathMetric[2];
	int **Path[2];
	int MinPathMetricIndex;

	/* Number of bits in every state = PRImpResLength-1 (= Order of PR polynomial) */
	/* Number of states in trellis = 2^(Order of PR polynomial) */
	NumStates = (int)pow(2,(PRImpResLength-1));

	/* 
	Path[0][i][j] : Bit at a delay (Delay-j) of the survivor path corresponding to state i
					at the previous time.
	Path[1][i][j] : Bit at a delay (Delay-j) of the survivor path corresponding to state i
					at the current time.
	*/
	Path[0] = (int**)malloc(NumStates*sizeof(int));
	Path[1] = (int**)malloc(NumStates*sizeof(int));
	for(i=0;i<NumStates;i++)
	{
		Path[0][i] = (int*)malloc(Delay*sizeof(int));
		Path[1][i] = (int*)malloc(Delay*sizeof(int));
	}
	for(i=0;i<NumStates;i++)
	{
		for(j=0;j<Delay;j++)
		{
			Path[0][i][j] = 0;
			Path[1][i][j] = 0;
		}
	}
	PathMetric[0] = (double *)malloc(NumStates * sizeof(double));
	PathMetric[1] = (double *)malloc(NumStates * sizeof(double));
	/* PathMetric[i] => The path metric of the valid output of state i at current time */
	/*
	NOTE that Pathmetric is initialized to 0 in the hope that by the time 'Delay' bits are
	transmitted, all of these are overwritten. If not, then when making a decision on a bit,
	the state for which Pathmetric is still 0 will be chosen, which is wrong. It cannot be
	initialized to a large number to avoid this problem because it is additive.
	*/
	for(i=0;i<NumStates;i++)
	{
		PathMetric[0][i] = 1e50;
		PathMetric[1][i] = 1e50;
	}
	PathMetric[0][0] = 0;
	PathMetric[1][0] = 0;
	
	/**** Main VITERBI loop ****/
	for(k=0;k<SectorLength;k++)
	{/* Loop over PaddedEqOutput */

		for(i=0;i<NumStates;i++)
		{/* Loop over all states */

			/*** For each state, calculating the path metric of the two incoming paths ***/

			/*
			If i = [i1 i2 ... in] (the array elements are binary) is the current state
			then its corresponding previous states are [0 i1 i2.....in-1] and [1 i1 i2....in-1].
			These are determined by	bit operations given below.
			*/

			/**
			Calculating the sample value and metric at current time for the transition from
			state [0 i1 i2....in-1]	to state [i1 i2....in].
			Corresponding Previous state is given by: (CurrentState >> 1) => shifting all
			bits to the right by 1 position and appending a ZERO at the left most position.
			**/
			sample0 = 0.0;
			for(j=0;j<PRImpResLength-1;j++)
			{
				sample0 += PRImpRes[j]*((i >> j) & 1);
			}
			sample0 += PRImpRes[PRImpResLength-1]*0;	/* Eventhough it is 0, it is explicitly written for clarity */
			sample0 = sample0*2;
			metric0 = PathMetric[0][i >> 1] + pow((EqualizedChannelOutput[k] - sample0),2);

			/**
			Calculating the sample value and metric at current time for the transition from
			state [1 i1 i2....in-1]	to state [i1 i2....in]
			Corresponding Previous state is given by: (CurrentState >> 1) | (1 << (CurrentStateLength-1))
			=> shifting all bits to the right by 1 position and appending a ONE at the
			left most position
			**/
			sample1 = 0.0;
			for(j=0;j<PRImpResLength-1;j++)
			{
				sample1 += PRImpRes[j]*((i >> j) & 1);
			}
			sample1 += PRImpRes[PRImpResLength-1]*1;
			sample1 = sample1*2;
			metric1 = PathMetric[0][(i >> 1)|(1 << (PRImpResLength-2))] + pow((EqualizedChannelOutput[k] - sample1),2);

			if(metric0 <= metric1)
			{
				/*
				New path is from the state [0 i1 i2...in-1], i.e current valid path for state i is
				the valid path of state [0 i1...in-1] in the previous time appended with bit 'in'
				*/
				/*
				(Path[0][i >> 1] << 1) => shifting all bits of path of the state [0 i1...in-1]
				to the left by 1 position.
				(i & 1) => The last bit of state i.
				(Path[0][i >> 1] << 1) | (i & 1) => appending last bit of i to the left-shifted path.
				*/
				for(m=0;m<Delay-1;m++)
				{
					Path[1][i][m] = Path[0][i >> 1][m+1];
				}
				Path[1][i][Delay-1] = i & 1;
				PathMetric[1][i] = metric0;
			}else{
				/*
				New path is from the state [1 i1 i2...in-1], i.e current valid path for state i is
				the valid path of state [1 i1...in-1] in the previous time appended with bit 'in'
				*/
				/* Bit operations are done as before */
				for(m=0;m<Delay-1;m++)
				{
					Path[1][i][m] = Path[0][(i >> 1)|(1 << (PRImpResLength-2))][m+1];
				}
				Path[1][i][Delay-1] = i & 1;
				PathMetric[1][i] = metric1;
			}
		}/* Loop over all states */

		/* Make a decision on the bit transmitted 'Delay' bit periods before */
		if(k >= Delay-1)
		{
			/** Deciding on the transmitted bit based on detection delay **/
			/* Finding the minimum path metric at the current time */
			MinPathMetricIndex = 0;
			for(i=1;i<NumStates;i++)
			{
				if(PathMetric[1][i] < PathMetric[1][MinPathMetricIndex])
					MinPathMetricIndex = i;
			}
			/* Detected output = first bit in Path[1][MinPathMetricIndex] */
			/* Path[1][MinPathMetricIndex] has 'Delay' number of bits */
			DetectedOutput[k-(Delay-1)] = Path[1][MinPathMetricIndex][0];
		}

		/* Exchange old and new Path and PathMetric values */
		for(i=0;i<NumStates;i++)
		{
			for(m=0;m<Delay;m++)
			{
				Path[0][i][m] = Path[1][i][m];
				Path[1][i][m] = 0;
			}
			PathMetric[0][i] = PathMetric[1][i];
			PathMetric[1][i] = 0.0;
		}
	}/* Loop over PaddedEqOutput */

	/* Detecting the last 'delay-1' bits */
	for(i=2;i<=Delay;i++)
	{
		DetectedOutput[SectorLength-1-(Delay-i)] = Path[0][MinPathMetricIndex][i-1];
	}
	for(i=0;i<NumStates;i++)
	{
		free(Path[0][i]);
		free(Path[1][i]);
	}
	free(Path[0]);
	free(Path[1]);
	free(PathMetric[0]);
	free(PathMetric[1]);
}


/*------------- SOFT-OUTPUT VITERBI ALGORITHM (SOVA) (DIFFERENT IMPLEMENTATION) ------------*/

void ClassicalSOVA(int Delay, double *EqualizedChannelOutput, int SectorLength, int *DetectedHardOutput, double *DetectedSoftOutput, double *PRImpRes, int PRImpResLength, double NoiseSigma)
{
	int i,j,k,m,NumStates;
	double metric0,metric1;
	double sample0,sample1;
	int state0,state1,survivingstate,MinPathMetricIndex;
	double *PathMetric[2];
	int **Path[2];
	int *VS[2];
	double  **ProbWrongDet[2];
	double NoiseVariance = NoiseSigma*NoiseSigma;
	double Delta = 0.0;
	double ProbWrongSurv = 0.0;
	int *DifferingBits;
	int VSLengthPrevTime=1,VSLengthCurrentTime=0;

	/* Number of bits in every state = PRImpResLength-1 */
	/* Number of states in trellis = 2^(Order of PR polynomial) */
	NumStates = (int)pow(2,(PRImpResLength-1));

	/* 
	Path[0][i][j] : Bit at a delay (Delay-j) of the survivor path corresponding to state i
					at the previous time.
	Path[1][i][j] : Bit at a delay (Delay-j) of the survivor path corresponding to state i
					at the current time.
	*/
	Path[0] = (int **)malloc(NumStates*sizeof(int));
	Path[1] = (int **)malloc(NumStates*sizeof(int));
	for(i=0;i<NumStates;i++)
	{
		Path[0][i] = (int*)malloc(Delay*sizeof(int));
		Path[1][i] = (int*)malloc(Delay*sizeof(int));
	}
	for(i=0;i<NumStates;i++)
	{
		for(j=0;j<Delay;j++)
		{
			Path[0][i][j] = 0;
			Path[1][i][j] = 0;
		}
	}

	PathMetric[0] = (double *)malloc(NumStates * sizeof(double));
	PathMetric[1] = (double *)malloc(NumStates * sizeof(double));
	/* PathMetric[i] => The path metric of the valid output of state i at current time */
	/*
	NOTE that Pathmetric is initialized to 0 in the hope that by the time 'Delay' bits are
	transmitted, all of these are overwritten. If not, then when making a decision on a bit,
	the state for which Pathmetric is still 0 will be chosen, which is wrong. It cannot be
	initialized to a large number to avoid this problem because it is additive.
	*/
	for(i=0;i<NumStates;i++)
	{
		PathMetric[0][i] = 0.0;
		PathMetric[1][i] = 0.0;
	}

	/*
	ProbWrongDet variables are used to store the probabilities that each bit is detected
	wrongly. As before, 0 indicates previous time and 1 indicates current time
	*/
	ProbWrongDet[0] = (double **)malloc(NumStates*sizeof(double));
	ProbWrongDet[1] = (double **)malloc(NumStates*sizeof(double));
	for(i=0;i<NumStates;i++){
		ProbWrongDet[0][i] = (double *)malloc(Delay*sizeof(double));
		ProbWrongDet[1][i] = (double *)malloc(Delay*sizeof(double));
	}
	for(i=0;i<NumStates;i++)
	{
		for(j=0;j<Delay;j++)
		{
			ProbWrongDet[0][i][j] = 0.0;
			ProbWrongDet[1][i][j] = 0.0;
		}
	}

	/* Used to calculate bits that differ between the two paths that converge to a state */
	DifferingBits = (int*)malloc(Delay*sizeof(int));
	for(i=0;i<Delay;i++)
	{
		DifferingBits[i]=0;
	}

	/* Initially, until time (PRImpResLength-1), all states are not valid */
	/* VS[0] => Valid States in the previous time */
	/* VS[1] => Valid States in the current time */
	VS[0] = (int *)malloc(sizeof(int));
	VS[1] = (int *)malloc(sizeof(int));
	VS[0][0]=0;	/* Assuming that we start with state 0 at time 0 */
	VS[1][0]=0;

	/**** Main SOVA loop ****/
	for(k=0;k<SectorLength;k++)
	{/* Loop over PaddedEqOutput */

		for(i=0;i<NumStates;i++)
		{/* Loop over all states */

			/*** For each state, calculating the path metric of the two incoming paths ***/

			/*
			If i = [i1 i2 ... in] (the array elements are binary) is the current state
			then its corresponding previous states are [0 i1 i2.....in-1] and [1 i1 i2....in-1].
			These are determined by	doing the bit operations as given below.
			*/

			/**
			Calculating the sample value and metric at current time for the transition from
			state [0 i1 i2....in-1]	to state [i1 i2....in].
			Corresponding Previous state is given by: (CurrentState >> 1) => shifting all
			bits to the right by 1 position and appending a ZERO at the left most position.
			**/
			sample0 = 0.0;
			for(j=0;j<PRImpResLength-1;j++)
			{
				sample0 += PRImpRes[j]*((i >> j) & 1);
			}
			sample0 += PRImpRes[PRImpResLength-1]*0;	/* Eventhough it is 0, it is explicitly written for clarity */
			sample0 = sample0*2;
			metric0 = PathMetric[0][i >> 1] + pow((EqualizedChannelOutput[k] - sample0),2);

			/**
			Calculating the sample value and metric at current time for the transition from
			state [1 i1 i2....in-1]	to state [i1 i2....in]
			Corresponding Previous state is given by: (CurrentState >> 1) | (1 << (CurrentStateLength-1))
			=> shifting all bits to the right by 1 position and appending a ONE at the
			left most position
			**/
			sample1 = 0.0;
			for(j=0;j<PRImpResLength-1;j++)
			{
				sample1 += PRImpRes[j]*((i >> j) & 1);
			}
			sample1 += PRImpRes[PRImpResLength-1]*1;
			sample1 = sample1*2;
			metric1 = PathMetric[0][(i >> 1)|(1 << (PRImpResLength-2))] + pow((EqualizedChannelOutput[k] - sample1),2);

			state0 = i >> 1;
			state1 = (i >> 1)|(1 << (PRImpResLength-2));

			/* 
			Assuming that we start with state 0, then until time (PRImpResLength-1),
			all states will have less than 2 valid incoming paths. For example, at time 1, the 
			incoming paths for any state is valid only if it comes from state 0, because we 
			assume that we start from state 0 in time 0.
			*/
			if(k < (PRImpResLength-1))
			{
				for(j=0;j<VSLengthPrevTime;j++)
				{
					if((i >> 1)==VS[0][j])
					{	
						/* This is the only valid incoming path */
						for(m=0;m<Delay-1;m++)
						{
							Path[1][i][m] = Path[0][i >> 1][m+1];
						}
						Path[1][i][Delay-1] = i & 1;
						PathMetric[1][i] = metric0;
						VSLengthCurrentTime++;
						VS[1] = (int*)realloc(VS[1],(VSLengthCurrentTime)*sizeof(int));
						/* 
						This state now becomes valid, i.e. any path emerging out of
						this state is valid in the next time
						*/
						VS[1][VSLengthCurrentTime-1] = i;
						break;	/* At most only one path is valid */
					}
					if(((i >> 1) | (1 << (PRImpResLength-2))) == VS[0][j])
					{
						/* This is the only valid incoming path */
						for(m=0;m<Delay-1;m++)
						{
							Path[1][i][m] = Path[0][(i >> 1)|(1 << (PRImpResLength-2))][m+1];
						}
						Path[1][i][Delay-1] = i & 1;
						PathMetric[1][i] = metric1;
						VSLengthCurrentTime++;
						VS[1] = (int*)realloc(VS[1],(VSLengthCurrentTime)*sizeof(int));
						/* 
						This state now becomes valid, i.e. any path emerging out of
						this state is valid in the next time
						*/
						VS[1][VSLengthCurrentTime-1] = i;
						break;	/* At most only one path is valid */
					}
				}
			}
			else
			{ /* Loop: All states and all incoming paths are valid */
				if(metric0 <= metric1)
				{
					/*
					New path is from the state [0 i1 i2...in-1], i.e current valid path for state i is
					the valid path of state [0 i1...in-1] in the previous time appended with bit 'in'
					*/

					/*
					(Path[0][i >> 1] << 1) => shifting all bits of path of the state [0 i1...in-1]
					to the left by 1 position.
					(i & 1) => The last bit of state i.
					(Path[0][i >> 1] << 1) | (i & 1) => appending last bit of i to the left-shifted path.
					*/
					for(m=0;m<Delay-1;m++)
					{
						Path[1][i][m] = Path[0][i >> 1][m+1];
					}
					Path[1][i][Delay-1] = i & 1;
					PathMetric[1][i] = metric0;
					survivingstate = state0;

				}else{
					/*
					New path is from the state [1 i1 i2...in-1], i.e current valid path for state i is
					the valid path of state [1 i1...in-1] in the previous time appended with bit 'in'
					*/
					/* Bit operations are done as before */
					for(m=0;m<Delay-1;m++)
					{
						Path[1][i][m] = Path[0][(i >> 1)|(1 << (PRImpResLength-2))][m+1];
					}
					Path[1][i][Delay-1] = i & 1;
					PathMetric[1][i] = metric1;
					survivingstate = state1;
				}

				/** Update Probability of wrong detection of bits **/
				Delta = fabs(metric0 - metric1)/(2*NoiseVariance);
				ProbWrongSurv = 1/(1+exp(Delta));

				/* Calculate the bits where the two incoming paths to state i differ, using XOR */
				for(j=0;j<Delay;j++)
				{
					DifferingBits[j] = (Path[0][state0][j] == Path[0][state1][j]) ? 0 : 1;
				}

				/* Obviously, they differ at the current time */
				ProbWrongDet[1][i][Delay-1] = ProbWrongSurv;

				/* For all bits, calculate Probablity of wrong detection */
				/*
				Note that we do not consider the bit in the path 'Delay' time period away (i.e 0th element in array),
				because it has already been detected in the previous iteration and will be overwritten in this iteration.
				*/
				for(j=0;j<Delay-1;j++)
				{
					/* Bits between the two paths differ */
					if(DifferingBits[j+1] == 1)
					{
						ProbWrongDet[1][i][Delay-2-j] = ProbWrongDet[0][survivingstate][Delay-1-j]*(1-ProbWrongSurv) + (1-ProbWrongDet[0][survivingstate][Delay-1-j])*ProbWrongSurv;
					}
					/* Bits between the two paths dont differ */
					else
					{
						ProbWrongDet[1][i][Delay-2-j] = ProbWrongDet[0][survivingstate][Delay-1-j];
					}
				}
			}/* Loop: All states and all incoming paths are valid */

		}/* Loop over all states */

		/* Make a decision on the bit transmitted 'Delay' bit periods before */
		if(k >= Delay-1)
		{
			/** Deciding on the transmitted bit based on detection delay **/
			/* Finding the minimum path metric at the current time */
			MinPathMetricIndex = 0;
			for(i=1;i<NumStates;i++){
				if(PathMetric[1][i] < PathMetric[1][MinPathMetricIndex])
					MinPathMetricIndex = i;
			}
			/* Detected output = first bit in Path[1][MinPathMetricIndex] */
			/* Path[1][MinPathMetricIndex] has 'Delay' number of bits */
			DetectedHardOutput[k-(Delay-1)] = Path[1][MinPathMetricIndex][0];

			/* Detecting soft output */
			if(DetectedHardOutput[k-(Delay-1)] == 1)
				DetectedSoftOutput[k-(Delay-1)] = ProbWrongDet[1][MinPathMetricIndex][0];
			else
				DetectedSoftOutput[k-(Delay-1)] = 1 - ProbWrongDet[1][MinPathMetricIndex][0];
		}

		/* Exchange old and new Path and PathMetric values */
		for(i=0;i<NumStates;i++)
		{
			for(m=0;m<Delay;m++)
			{
				Path[0][i][m] = Path[1][i][m];
				Path[1][i][m] = 0;
			}
			PathMetric[0][i] = PathMetric[1][i];
			PathMetric[1][i] = 0.0;
		}

		for(i=0;i<NumStates;i++)
		{
			for(j=0;j<Delay;j++)
			{
				ProbWrongDet[0][i][j] = ProbWrongDet[1][i][j];
				ProbWrongDet[1][i][j] = 0.0;
			}
		}

		/* Exchange old and new Valid State information */
		if(k < (PRImpResLength - 1))
		{
			VS[0] = (int*)realloc(VS[0],VSLengthCurrentTime*sizeof(int));
			for(i=0;i<VSLengthCurrentTime;i++)
			{
				VS[0][i] = VS[1][i];
				VS[1][i] = 0;
			}
			VSLengthPrevTime = VSLengthCurrentTime;
			VSLengthCurrentTime = 0;
		}

	}/* Loop over PaddedEqOutput */
	
	/* Detecting the last 'delay-1' bits */
	/* Current time information has been copied to the old time information, 
	so remember to use the old info.
	*/
	for(i=2;i<=Delay;i++)
	{
		DetectedHardOutput[SectorLength-1-(Delay-i)] = Path[0][MinPathMetricIndex][i-1];
		/* Detecting soft output */
		if(DetectedHardOutput[SectorLength-1-(Delay-i)] == 1)
			DetectedSoftOutput[SectorLength-1-(Delay-i)] = ProbWrongDet[0][MinPathMetricIndex][i-1];
		else
			DetectedSoftOutput[SectorLength-1-(Delay-i)] = 1 - ProbWrongDet[0][MinPathMetricIndex][i-1];
	}
	free(Path[0]);
	free(Path[1]);
	free(VS[0]);
	free(VS[1]);
	free(PathMetric[0]);
	free(PathMetric[1]);
	free(DifferingBits);
	for(i=0;i<NumStates;i++)
	{
		free(ProbWrongDet[0][i]);
		free(ProbWrongDet[1][i]);
	}
	free(ProbWrongDet[0]);
	free(ProbWrongDet[1]);
}


/*---------------------------- SOVA FOR 8/9 MTR(2/3;11) CODE ----------------------------*/

void SOVA_8By9MTRCode(int Delay, double *EqualizedChannelOutput, int PrePaddingLength, int SectorLength, int *DetectedHardOutput, double *DetectedSoftOutput, double *PRImpRes, int PRImpResLength, double NoiseSigma)
{
	int i,j,k,m,NumStates, NumCodeword;
	double metric0,metric1;
	double sample0,sample1;
	int state0,state1,survivingstate,MinPathMetricIndex;
	double *PathMetric[2];
	int **Path[2];
	int *VS[2];
	double  **ProbWrongDet[2];
	double NoiseVariance = NoiseSigma*NoiseSigma;
	double Delta = 0.0;
	double ProbWrongSurv = 0.0;
	int *DifferingBits;
	int VSLengthPrevTime=1,VSLengthCurrentTime=0;

	/* Number of bits in every state = PRImpResLength-1 */
	/* Number of states in trellis = 2^(Order of PR polynomial) */
	NumStates = (int)pow(2,(PRImpResLength-1));

	/* 
	Path[0][i][j] : Bit at a delay (Delay-j) of the survivor path corresponding to state i
					at the previous time.
	Path[1][i][j] : Bit at a delay (Delay-j) of the survivor path corresponding to state i
					at the current time.
	*/
	Path[0] = (int **)malloc(NumStates*sizeof(int));
	Path[1] = (int **)malloc(NumStates*sizeof(int));
	for(i=0;i<NumStates;i++)
	{
		Path[0][i] = (int*)malloc(Delay*sizeof(int));
		Path[1][i] = (int*)malloc(Delay*sizeof(int));
	}
	for(i=0;i<NumStates;i++)
	{
		for(j=0;j<Delay;j++)
		{
			Path[0][i][j] = 0;
			Path[1][i][j] = 0;
		}
	}

	PathMetric[0] = (double *)malloc(NumStates * sizeof(double));
	PathMetric[1] = (double *)malloc(NumStates * sizeof(double));
	/* PathMetric[i] => The path metric of the valid output of state i at current time */
	/*
	NOTE that Pathmetric is initialized to 0 in the hope that by the time 'Delay' bits are
	transmitted, all of these are overwritten. If not, then when making a decision on a bit,
	the state for which Pathmetric is still 0 will be chosen, which is wrong. It cannot be
	initialized to a large number to avoid this problem because it is additive.
	*/
	for(i=0;i<NumStates;i++)
	{
		PathMetric[0][i] = 0.0;
		PathMetric[1][i] = 0.0;
	}

	/*
	ProbWrongDet variables are used to store the probabilities that each bit is detected
	wrongly. As before, 0 indicates previous time and 1 indicates current time
	*/
	ProbWrongDet[0] = (double **)malloc(NumStates*sizeof(double));
	ProbWrongDet[1] = (double **)malloc(NumStates*sizeof(double));
	for(i=0;i<NumStates;i++){
		ProbWrongDet[0][i] = (double *)malloc(Delay*sizeof(double));
		ProbWrongDet[1][i] = (double *)malloc(Delay*sizeof(double));
	}
	for(i=0;i<NumStates;i++)
	{
		for(j=0;j<Delay;j++)
		{
			ProbWrongDet[0][i][j] = 0.0;
			ProbWrongDet[1][i][j] = 0.0;
		}
	}

	/* Used to calculate bits that differ between the two paths that converge to a state */
	DifferingBits = (int*)malloc(Delay*sizeof(int));
	for(i=0;i<Delay;i++)
	{
		DifferingBits[i]=0;
	}

	/* Initially, until time (PRImpResLength-1), all states are not valid */
	/* VS[0] => Valid States in the previous time */
	/* VS[1] => Valid States in the current time */
	VS[0] = (int *)malloc(sizeof(int));
	VS[1] = (int *)malloc(sizeof(int));
	VS[0][0]=0;	/* Assuming that we start with state 0 at time 0 */
	VS[1][0]=0;

	/** Adapt Viterbi according to MTR constraints **/
	/* Check if the current state is a valid state for the 6/7 MTR(2;8) code */
	if(PRImpResLength-1 == 2)
	{
		printf("\nERROR: -> Viterbi() : Viterbi not defined for PR Impulse response length = 3 \n");
		exit(-1);

	}else if(PRImpResLength-1 == 3)
	{
		/* All states are valid */
	}else if(PRImpResLength-1 == 4)
	{
		PathMetric[0][5] = 1e50;	/* Some large number, so that this path will never be selected */
		PathMetric[0][10] = 1e50;	/* Some large number, so that this path will never be selected */

	}else if(PRImpResLength-1 == 5)
	{
		printf("\nERROR: -> SOVA() : SOVA not defined for PR impulse response length > 6 \n");
		exit(-1);
		//if(i==5 || i==10 || i==11 || i==20 || i==21 || i==36)
		//	PathMetric[1][i] = 1e50;
	}
	else
	{
		printf("\nERROR: -> Viterbi() : Viterbi not defined for PR impulse response length > 6 \n");
		exit(-1);
	}

	/**** Main SOVA loop ****/
	for(k=0;k<SectorLength;k++)
	{/* Loop over PaddedEqOutput */

		for(i=0;i<NumStates;i++)
		{/* Loop over all states */

			/*** For each state, calculating the path metric of the two incoming paths ***/

			/*
			If i = [i1 i2 ... in] (the array elements are binary) is the current state
			then its corresponding previous states are [0 i1 i2.....in-1] and [1 i1 i2....in-1].
			These are determined by	doing the bit operations as given below.
			*/

			/**
			Calculating the sample value and metric at current time for the transition from
			state [0 i1 i2....in-1]	to state [i1 i2....in].
			Corresponding Previous state is given by: (CurrentState >> 1) => shifting all
			bits to the right by 1 position and appending a ZERO at the left most position.
			**/
			sample0 = 0.0;
			for(j=0;j<PRImpResLength-1;j++)
			{
				sample0 += PRImpRes[j]*((i >> j) & 1);
			}
			sample0 += PRImpRes[PRImpResLength-1]*0;	/* Eventhough it is 0, it is explicitly written for clarity */
			sample0 = sample0*2;
			metric0 = PathMetric[0][i >> 1] + pow((EqualizedChannelOutput[k] - sample0),2);

			/**
			Calculating the sample value and metric at current time for the transition from
			state [1 i1 i2....in-1]	to state [i1 i2....in]
			Corresponding Previous state is given by: (CurrentState >> 1) | (1 << (CurrentStateLength-1))
			=> shifting all bits to the right by 1 position and appending a ONE at the
			left most position
			**/
			sample1 = 0.0;
			for(j=0;j<PRImpResLength-1;j++)
			{
				sample1 += PRImpRes[j]*((i >> j) & 1);
			}
			sample1 += PRImpRes[PRImpResLength-1]*1;
			sample1 = sample1*2;
			metric1 = PathMetric[0][(i >> 1)|(1 << (PRImpResLength-2))] + pow((EqualizedChannelOutput[k] - sample1),2);

			state0 = i >> 1;
			state1 = (i >> 1)|(1 << (PRImpResLength-2));

			/* 
			Assuming that we start with state 0, then until time (PRImpResLength-1),
			all states will have less than 2 valid incoming paths. For example, at time 1, the 
			incoming paths for any state is valid only if it comes from state 0, because we 
			assume that we start from state 0 in time 0.
			*/
			if(k < (PRImpResLength-1))
			{
				for(j=0;j<VSLengthPrevTime;j++)
				{
					if((i >> 1)==VS[0][j])
					{	
						/* This is the only valid incoming path */
						for(m=0;m<Delay-1;m++)
						{
							Path[1][i][m] = Path[0][i >> 1][m+1];
						}
						Path[1][i][Delay-1] = i & 1;
						PathMetric[1][i] = metric0;
						VSLengthCurrentTime++;
						VS[1] = (int*)realloc(VS[1],(VSLengthCurrentTime)*sizeof(int));
						/* 
						This state now becomes valid, i.e. any path emerging out of
						this state is valid in the next time
						*/
						VS[1][VSLengthCurrentTime-1] = i;
						break;	/* At most only one path is valid */
					}
					if(((i >> 1) | (1 << (PRImpResLength-2))) == VS[0][j])
					{
						/* This is the only valid incoming path */
						for(m=0;m<Delay-1;m++)
						{
							Path[1][i][m] = Path[0][(i >> 1)|(1 << (PRImpResLength-2))][m+1];
						}
						Path[1][i][Delay-1] = i & 1;
						PathMetric[1][i] = metric1;
						VSLengthCurrentTime++;
						VS[1] = (int*)realloc(VS[1],(VSLengthCurrentTime)*sizeof(int));
						/* 
						This state now becomes valid, i.e. any path emerging out of
						this state is valid in the next time
						*/
						VS[1][VSLengthCurrentTime-1] = i;
						break;	/* At most only one path is valid */
					}
				}
			}
			else
			{ /* Loop: All states and all incoming paths are valid */
				if(metric0 <= metric1)
				{
					/*
					New path is from the state [0 i1 i2...in-1], i.e current valid path for state i is
					the valid path of state [0 i1...in-1] in the previous time appended with bit 'in'
					*/

					/*
					(Path[0][i >> 1] << 1) => shifting all bits of path of the state [0 i1...in-1]
					to the left by 1 position.
					(i & 1) => The last bit of state i.
					(Path[0][i >> 1] << 1) | (i & 1) => appending last bit of i to the left-shifted path.
					*/
					for(m=0;m<Delay-1;m++)
					{
						Path[1][i][m] = Path[0][i >> 1][m+1];
					}
					Path[1][i][Delay-1] = i & 1;
					PathMetric[1][i] = metric0;
					survivingstate = state0;

				}else{
					/*
					New path is from the state [1 i1 i2...in-1], i.e current valid path for state i is
					the valid path of state [1 i1...in-1] in the previous time appended with bit 'in'
					*/
					/* Bit operations are done as before */
					for(m=0;m<Delay-1;m++)
					{
						Path[1][i][m] = Path[0][(i >> 1)|(1 << (PRImpResLength-2))][m+1];
					}
					Path[1][i][Delay-1] = i & 1;
					PathMetric[1][i] = metric1;
					survivingstate = state1;
				}

				/** Update Probability of wrong detection of bits **/
				Delta = fabs(metric0 - metric1)/(2*NoiseVariance);
				ProbWrongSurv = 1/(1+exp(Delta));

				/* Calculate the bits where the two incoming paths to state i differ, using XOR */
				for(j=0;j<Delay;j++)
				{
					DifferingBits[j] = (Path[0][state0][j] == Path[0][state1][j]) ? 0 : 1;
				}

				/* Obviously, they differ at the current time */
				ProbWrongDet[1][i][Delay-1] = ProbWrongSurv;

				/* For all bits, calculate Probablity of wrong detection */
				/*
				Note that we do not consider the bit in the path 'Delay' time period away (i.e 0th element in array),
				because it has already been detected in the previous iteration and will be overwritten in this iteration.
				*/
				for(j=0;j<Delay-1;j++)
				{
					/* Bits between the two paths differ */
					if(DifferingBits[j+1] == 1)
					{
						ProbWrongDet[1][i][Delay-2-j] = ProbWrongDet[0][survivingstate][Delay-1-j]*(1-ProbWrongSurv) + (1-ProbWrongDet[0][survivingstate][Delay-1-j])*ProbWrongSurv;
					}
					/* Bits between the two paths dont differ */
					else
					{
						ProbWrongDet[1][i][Delay-2-j] = ProbWrongDet[0][survivingstate][Delay-1-j];
					}
				}
			}/* Loop: All states and all incoming paths are valid */

			/* 
			Check for state validity for the data sequence.
			Note that we also ignore the first bit of the data sequence apart from
			the padded bits. This is because of the NRZI-NRZ conversion.
			*/	
			if((k>=PrePaddingLength+1) && (k<(PrePaddingLength+SectorLength)))
			{
				/* Count the number of codewords */
				NumCodeword = (k-PrePaddingLength-1)/9 + 1;	/* Starting from 1 */

				/* Check if the current state is a valid state for the 8/9 MTR code */
				if(PRImpResLength-1 == 2)
				{
					printf("\nERROR: -> Viterbi() : Viterbi not defined for PR Impulse response length = 3 \n");
					exit(-1);

				}else if(PRImpResLength-1 == 3)
				{
					/* All states are valid. But certain edges aren't, which is taken care elsewhere. */
				}else if(PRImpResLength-1 == 4)
				{
					/* For 8/9 MTR code, tribits are allowed only at odd bit locations. Since, the codeword length is 
					odd(9), valid states in a viterbi detector will change from codeword to codeword, apart from changing
					at every alternate times. That is, for the 1st codeword, tribits at odd time units will be allowed 
					but for the 2nd codeword tribits at even time units will be allowed.  Now, since 
					viterbi operates on NRZ bits, whose length is one more than the NRZI bits, allowed tribit locations
					are switched(even <-> odd). An another issue to worry about is the length of the prepadding bits. 
					If it is even, then as mentioned above, for NRZ bits, tribits locations are switched. 
					If it is odd, then for NRZ bits, tribit locations remain the same as for NRZI. Here, it is assumed that
					prepadding bit length is even. To summarise (assuming even prepadding length):

					NRZI: 
						@ Odd Codewords(starting from 1): Tribits at EVEN locations allowed
						@ Even Codewords(starting from 1): Tribits at ODD locations allowed

					NRZ: 
						@ Odd Codewords(starting from 1): Tribits at ODD locations allowed
						@ Even Codewords(starting from 1): Tribits at EVEN locations allowed
					*/

					/* Detecting Odd NRZI bits (or Even NRZ bits) */
					if(((k-PrePaddingLength)-(NumCodeword-1)*9)%2 == 1)// || ((k-PrePaddingLength)-(NumCodeword-1)*9) == 2)
					{
						if(i==5 || i==10)				/* Invalid states: 0101 & 1010 */
							PathMetric[1][i] = 1e50;	/* Some large number, so that this path will never be selected */
					}else
					{
						/* All states are valid */

					}
				}else if(PRImpResLength-1 == 5)
				{

					printf("\nERROR: -> Viterbi() : Viterbi not defined for PR impulse response length = 6 \n");
					exit(-1);
				}
				else
				{
					printf("\nERROR: -> Viterbi() : Viterbi not defined for PR impulse response length > 6 \n");
					exit(-1);
				}
			}

		}/* Loop over all states */

		/* Make a decision on the bit transmitted 'Delay' bit periods before */
		if(k >= Delay-1)
		{
			/** Deciding on the transmitted bit based on detection delay **/
			/* Finding the minimum path metric at the current time */
			MinPathMetricIndex = 0;
			for(i=1;i<NumStates;i++){
				if(PathMetric[1][i] < PathMetric[1][MinPathMetricIndex])
					MinPathMetricIndex = i;
			}
			/* Detected output = first bit in Path[1][MinPathMetricIndex] */
			/* Path[1][MinPathMetricIndex] has 'Delay' number of bits */
			DetectedHardOutput[k-(Delay-1)] = Path[1][MinPathMetricIndex][0];

			/* Detecting soft output */
			if(DetectedHardOutput[k-(Delay-1)] == 1)
				DetectedSoftOutput[k-(Delay-1)] = ProbWrongDet[1][MinPathMetricIndex][0];
			else
				DetectedSoftOutput[k-(Delay-1)] = 1 - ProbWrongDet[1][MinPathMetricIndex][0];
		}

		/* Exchange old and new Path and PathMetric values */
		for(i=0;i<NumStates;i++)
		{
			for(m=0;m<Delay;m++)
			{
				Path[0][i][m] = Path[1][i][m];
				Path[1][i][m] = 0;
			}
			PathMetric[0][i] = PathMetric[1][i];
			PathMetric[1][i] = 0.0;
		}

		for(i=0;i<NumStates;i++)
		{
			for(j=0;j<Delay;j++)
			{
				ProbWrongDet[0][i][j] = ProbWrongDet[1][i][j];
				ProbWrongDet[1][i][j] = 0.0;
			}
		}

		/* Exchange old and new Valid State information */
		if(k < (PRImpResLength - 1))
		{
			VS[0] = (int*)realloc(VS[0],VSLengthCurrentTime*sizeof(int));
			for(i=0;i<VSLengthCurrentTime;i++)
			{
				VS[0][i] = VS[1][i];
				VS[1][i] = 0;
			}
			VSLengthPrevTime = VSLengthCurrentTime;
			VSLengthCurrentTime = 0;
		}

	}/* Loop over PaddedEqOutput */
	
	/* Detecting the last 'delay-1' bits */
	/* Current time information has been copied to the old time information, 
	so remember to use the old info.
	*/
	for(i=2;i<=Delay;i++)
	{
		DetectedHardOutput[SectorLength-1-(Delay-i)] = Path[0][MinPathMetricIndex][i-1];
		/* Detecting soft output */
		if(DetectedHardOutput[SectorLength-1-(Delay-i)] == 1)
			DetectedSoftOutput[SectorLength-1-(Delay-i)] = ProbWrongDet[0][MinPathMetricIndex][i-1];
		else
			DetectedSoftOutput[SectorLength-1-(Delay-i)] = 1 - ProbWrongDet[0][MinPathMetricIndex][i-1];
	}
	free(Path[0]);
	free(Path[1]);
	free(VS[0]);
	free(VS[1]);
	free(PathMetric[0]);
	free(PathMetric[1]);
	free(DifferingBits);
	for(i=0;i<NumStates;i++)
	{
		free(ProbWrongDet[0][i]);
		free(ProbWrongDet[1][i]);
	}
	free(ProbWrongDet[0]);
	free(ProbWrongDet[1]);
}


/*---------------------------- SOVA FOR 6/7 MTR CODE ----------------------------*/

void SOVA_6By7MTRCode(int Delay, double *EqualizedChannelOutput, int PrePaddingLength, int SectorLength, int *DetectedHardOutput, double *DetectedSoftOutput, double *PRImpRes, int PRImpResLength, double NoiseSigma)
{
	int i,j,k,m,NumStates, NumCodeword;
	double metric0,metric1;
	double sample0,sample1;
	int state0,state1,survivingstate,MinPathMetricIndex;
	double *PathMetric[2];
	int **Path[2];
	int *VS[2];
	double  **ProbWrongDet[2];
	double NoiseVariance = NoiseSigma*NoiseSigma;
	double Delta = 0.0;
	double ProbWrongSurv = 0.0;
	int *DifferingBits;
	int VSLengthPrevTime=1,VSLengthCurrentTime=0;

	/* Number of bits in every state = PRImpResLength-1 */
	/* Number of states in trellis = 2^(Order of PR polynomial) */
	NumStates = (int)pow(2,(PRImpResLength-1));

	/* 
	Path[0][i][j] : Bit at a delay (Delay-j) of the survivor path corresponding to state i
					at the previous time.
	Path[1][i][j] : Bit at a delay (Delay-j) of the survivor path corresponding to state i
					at the current time.
	*/
	Path[0] = (int **)malloc(NumStates*sizeof(int));
	Path[1] = (int **)malloc(NumStates*sizeof(int));
	for(i=0;i<NumStates;i++)
	{
		Path[0][i] = (int*)malloc(Delay*sizeof(int));
		Path[1][i] = (int*)malloc(Delay*sizeof(int));
	}
	for(i=0;i<NumStates;i++)
	{
		for(j=0;j<Delay;j++)
		{
			Path[0][i][j] = 0;
			Path[1][i][j] = 0;
		}
	}

	PathMetric[0] = (double *)malloc(NumStates * sizeof(double));
	PathMetric[1] = (double *)malloc(NumStates * sizeof(double));
	/* PathMetric[i] => The path metric of the valid output of state i at current time */
	/*
	NOTE that Pathmetric is initialized to 0 in the hope that by the time 'Delay' bits are
	transmitted, all of these are overwritten. If not, then when making a decision on a bit,
	the state for which Pathmetric is still 0 will be chosen, which is wrong. It cannot be
	initialized to a large number to avoid this problem because it is additive.
	*/
	for(i=0;i<NumStates;i++)
	{
		PathMetric[0][i] = 0.0;
		PathMetric[1][i] = 0.0;
	}

	/*
	ProbWrongDet variables are used to store the probabilities that each bit is detected
	wrongly. As before, 0 indicates previous time and 1 indicates current time
	*/
	ProbWrongDet[0] = (double **)malloc(NumStates*sizeof(double));
	ProbWrongDet[1] = (double **)malloc(NumStates*sizeof(double));
	for(i=0;i<NumStates;i++){
		ProbWrongDet[0][i] = (double *)malloc(Delay*sizeof(double));
		ProbWrongDet[1][i] = (double *)malloc(Delay*sizeof(double));
	}
	for(i=0;i<NumStates;i++)
	{
		for(j=0;j<Delay;j++)
		{
			ProbWrongDet[0][i][j] = 0.0;
			ProbWrongDet[1][i][j] = 0.0;
		}
	}

	/* Used to calculate bits that differ between the two paths that converge to a state */
	DifferingBits = (int*)malloc(Delay*sizeof(int));
	for(i=0;i<Delay;i++)
	{
		DifferingBits[i]=0;
	}

	/* Initially, until time (PRImpResLength-1), all states are not valid */
	/* VS[0] => Valid States in the previous time */
	/* VS[1] => Valid States in the current time */
	VS[0] = (int *)malloc(sizeof(int));
	VS[1] = (int *)malloc(sizeof(int));
	VS[0][0]=0;	/* Assuming that we start with state 0 at time 0 */
	VS[1][0]=0;

	/** Adapt Viterbi according to MTR constraints **/
	/* Check if the current state is a valid state for the 6/7 MTR(2;8) code */
	if(PRImpResLength-1 == 2)
	{
		printf("\nERROR: -> Viterbi() : Viterbi not defined for PR Impulse response length = 3 \n");
		exit(-1);

	}else if(PRImpResLength-1 == 3)
	{
		/* All states are valid */
	}else if(PRImpResLength-1 == 4)
	{
		PathMetric[0][5] = 1e50;	/* Some large number, so that this path will never be selected */
		PathMetric[0][10] = 1e50;	/* Some large number, so that this path will never be selected */

	}else if(PRImpResLength-1 == 5)
	{
		printf("\nERROR: -> SOVA() : SOVA not defined for PR impulse response length > 6 \n");
		exit(-1);
		//if(i==5 || i==10 || i==11 || i==20 || i==21 || i==36)
		//	PathMetric[1][i] = 1e50;
	}
	else
	{
		printf("\nERROR: -> Viterbi() : Viterbi not defined for PR impulse response length > 6 \n");
		exit(-1);
	}

	/**** Main SOVA loop ****/
	for(k=0;k<SectorLength;k++)
	{/* Loop over PaddedEqOutput */

		for(i=0;i<NumStates;i++)
		{/* Loop over all states */

			/*** For each state, calculating the path metric of the two incoming paths ***/

			/*
			If i = [i1 i2 ... in] (the array elements are binary) is the current state
			then its corresponding previous states are [0 i1 i2.....in-1] and [1 i1 i2....in-1].
			These are determined by	doing the bit operations as given below.
			*/

			/**
			Calculating the sample value and metric at current time for the transition from
			state [0 i1 i2....in-1]	to state [i1 i2....in].
			Corresponding Previous state is given by: (CurrentState >> 1) => shifting all
			bits to the right by 1 position and appending a ZERO at the left most position.
			**/
			sample0 = 0.0;
			for(j=0;j<PRImpResLength-1;j++)
			{
				sample0 += PRImpRes[j]*((i >> j) & 1);
			}
			sample0 += PRImpRes[PRImpResLength-1]*0;	/* Eventhough it is 0, it is explicitly written for clarity */
			sample0 = sample0*2;
			metric0 = PathMetric[0][i >> 1] + pow((EqualizedChannelOutput[k] - sample0),2);

			/**
			Calculating the sample value and metric at current time for the transition from
			state [1 i1 i2....in-1]	to state [i1 i2....in]
			Corresponding Previous state is given by: (CurrentState >> 1) | (1 << (CurrentStateLength-1))
			=> shifting all bits to the right by 1 position and appending a ONE at the
			left most position
			**/
			sample1 = 0.0;
			for(j=0;j<PRImpResLength-1;j++)
			{
				sample1 += PRImpRes[j]*((i >> j) & 1);
			}
			sample1 += PRImpRes[PRImpResLength-1]*1;
			sample1 = sample1*2;
			metric1 = PathMetric[0][(i >> 1)|(1 << (PRImpResLength-2))] + pow((EqualizedChannelOutput[k] - sample1),2);

			state0 = i >> 1;
			state1 = (i >> 1)|(1 << (PRImpResLength-2));

			/* 
			Assuming that we start with state 0, then until time (PRImpResLength-1),
			all states will have less than 2 valid incoming paths. For example, at time 1, the 
			incoming paths for any state is valid only if it comes from state 0, because we 
			assume that we start from state 0 in time 0.
			*/
			if(k < (PRImpResLength-1))
			{
				for(j=0;j<VSLengthPrevTime;j++)
				{
					if((i >> 1)==VS[0][j])
					{	
						/* This is the only valid incoming path */
						for(m=0;m<Delay-1;m++)
						{
							Path[1][i][m] = Path[0][i >> 1][m+1];
						}
						Path[1][i][Delay-1] = i & 1;
						PathMetric[1][i] = metric0;
						VSLengthCurrentTime++;
						VS[1] = (int*)realloc(VS[1],(VSLengthCurrentTime)*sizeof(int));
						/* 
						This state now becomes valid, i.e. any path emerging out of
						this state is valid in the next time
						*/
						VS[1][VSLengthCurrentTime-1] = i;
						break;	/* At most only one path is valid */
					}
					if(((i >> 1) | (1 << (PRImpResLength-2))) == VS[0][j])
					{
						/* This is the only valid incoming path */
						for(m=0;m<Delay-1;m++)
						{
							Path[1][i][m] = Path[0][(i >> 1)|(1 << (PRImpResLength-2))][m+1];
						}
						Path[1][i][Delay-1] = i & 1;
						PathMetric[1][i] = metric1;
						VSLengthCurrentTime++;
						VS[1] = (int*)realloc(VS[1],(VSLengthCurrentTime)*sizeof(int));
						/* 
						This state now becomes valid, i.e. any path emerging out of
						this state is valid in the next time
						*/
						VS[1][VSLengthCurrentTime-1] = i;
						break;	/* At most only one path is valid */
					}
				}
			}
			else
			{ /* Loop: All states and all incoming paths are valid */
				if(metric0 <= metric1)
				{
					/*
					New path is from the state [0 i1 i2...in-1], i.e current valid path for state i is
					the valid path of state [0 i1...in-1] in the previous time appended with bit 'in'
					*/

					/*
					(Path[0][i >> 1] << 1) => shifting all bits of path of the state [0 i1...in-1]
					to the left by 1 position.
					(i & 1) => The last bit of state i.
					(Path[0][i >> 1] << 1) | (i & 1) => appending last bit of i to the left-shifted path.
					*/
					for(m=0;m<Delay-1;m++)
					{
						Path[1][i][m] = Path[0][i >> 1][m+1];
					}
					Path[1][i][Delay-1] = i & 1;
					PathMetric[1][i] = metric0;
					survivingstate = state0;

				}else{
					/*
					New path is from the state [1 i1 i2...in-1], i.e current valid path for state i is
					the valid path of state [1 i1...in-1] in the previous time appended with bit 'in'
					*/
					/* Bit operations are done as before */
					for(m=0;m<Delay-1;m++)
					{
						Path[1][i][m] = Path[0][(i >> 1)|(1 << (PRImpResLength-2))][m+1];
					}
					Path[1][i][Delay-1] = i & 1;
					PathMetric[1][i] = metric1;
					survivingstate = state1;
				}

				/** Update Probability of wrong detection of bits **/
				/* TEST CODE */
				if(k==0)
					printf("RUNNING TEST CODE -> SOVA_6By7()\n");
				Delta = fabs(metric0 - metric1)/(4*NoiseVariance);
				ProbWrongSurv = 1/(1+exp(Delta));

				/* Calculate the bits where the two incoming paths to state i differ */
				for(j=0;j<Delay;j++)
				{
					DifferingBits[j] = (Path[0][state0][j] == Path[0][state1][j]) ? 0 : 1;
				}

				/* Obviously, they differ at the current time */
				ProbWrongDet[1][i][Delay-1] = 0.0;//ProbWrongSurv;

				/* For all bits, calculate Probablity of wrong detection */
				/*
				Note that we do not consider the bit in the path 'Delay' time period away (i.e 0th element in array),
				because it has already been detected in the previous iteration and will be overwritten in this iteration.
				*/
				//for(j=0;j<Delay-1;j++)
				for(j=1;j<Delay;j++)
				{
					/* Bits between the two paths differ */
					if(DifferingBits[j] == 1)
					{
						//ProbWrongDet[1][i][Delay-2-j] = ProbWrongDet[0][survivingstate][Delay-1-j]*(1-ProbWrongSurv) + (1-ProbWrongDet[0][survivingstate][Delay-1-j])*ProbWrongSurv;
						ProbWrongDet[1][i][j-1] = ProbWrongDet[0][survivingstate][j]*(1-ProbWrongSurv) + (1-ProbWrongDet[0][survivingstate][j])*ProbWrongSurv;
					}
					/* Bits between the two paths dont differ */
					else
					{
						//ProbWrongDet[1][i][Delay-2-j] = ProbWrongDet[0][survivingstate][Delay-1-j];
						ProbWrongDet[1][i][j-1] = ProbWrongDet[0][survivingstate][j];
					}
				}
			}/* Loop: All states and all incoming paths are valid */

			/* 
			Check for state validity for the data sequence.
			Note that we also ignore the first bit of the data sequence apart from
			the padded bits. This is because of the NRZI-NRZ conversion.
			*/	
			if((k>=PrePaddingLength+1) && (k<(PrePaddingLength+SectorLength)))
			{
				/* Count the number of codewords */
				NumCodeword = (k-PrePaddingLength-1)/7 + 1;	/* Starting from 1 */

				/* Check if the current state is a valid state for the 8/9 MTR code */
				if(PRImpResLength-1 == 2)
				{
					printf("\nERROR: -> Viterbi() : Viterbi not defined for PR Impulse response length = 3 \n");
					exit(-1);

				}else if(PRImpResLength-1 == 3)
				{
					/* All states are valid. But certain edges aren't, which is taken care elsewhere. */

				}else if(PRImpResLength-1 == 4)
				{
					if(i==5 || i==10)				/* Invalid states: 0101 & 1010 */
						PathMetric[1][i] = 1e50;	/* Some large number, so that this path will never be selected */

				}else if(PRImpResLength-1 == 5)
				{

					printf("\nERROR: -> Viterbi() : Viterbi not defined for PR impulse response length = 6 \n");
					exit(-1);
				}
				else
				{
					printf("\nERROR: -> Viterbi() : Viterbi not defined for PR impulse response length > 6 \n");
					exit(-1);
				}
			}

		}/* Loop over all states */

		/* Make a decision on the bit transmitted 'Delay' bit periods before */
		if(k >= Delay-1)
		{
			/** Deciding on the transmitted bit based on detection delay **/
			/* Finding the minimum path metric at the current time */
			MinPathMetricIndex = 0;
			for(i=1;i<NumStates;i++){
				if(PathMetric[1][i] < PathMetric[1][MinPathMetricIndex])
					MinPathMetricIndex = i;
			}
			/* Detected output = first bit in Path[1][MinPathMetricIndex] */
			/* Path[1][MinPathMetricIndex] has 'Delay' number of bits */
			DetectedHardOutput[k-(Delay-1)] = Path[1][MinPathMetricIndex][0];

			/* Detecting soft output */
			//if(DetectedHardOutput[k-(Delay-1)] == 1)
				DetectedSoftOutput[k-(Delay-1)] = ProbWrongDet[1][MinPathMetricIndex][0];
			//else
			//	DetectedSoftOutput[k-(Delay-1)] = 1 - ProbWrongDet[1][MinPathMetricIndex][0];
		}

		/* Exchange old and new Path and PathMetric values */
		for(i=0;i<NumStates;i++)
		{
			for(m=0;m<Delay;m++)
			{
				Path[0][i][m] = Path[1][i][m];
				Path[1][i][m] = 0;
			}
			PathMetric[0][i] = PathMetric[1][i];
			PathMetric[1][i] = 0.0;
		}

		for(i=0;i<NumStates;i++)
		{
			for(j=0;j<Delay;j++)
			{
				ProbWrongDet[0][i][j] = ProbWrongDet[1][i][j];
				ProbWrongDet[1][i][j] = 0.0;
			}
		}

		/* Exchange old and new Valid State information */
		if(k < (PRImpResLength - 1))
		{
			VS[0] = (int*)realloc(VS[0],VSLengthCurrentTime*sizeof(int));
			for(i=0;i<VSLengthCurrentTime;i++)
			{
				VS[0][i] = VS[1][i];
				VS[1][i] = 0;
			}
			VSLengthPrevTime = VSLengthCurrentTime;
			VSLengthCurrentTime = 0;
		}

	}/* Loop over PaddedEqOutput */
	
	/* Detecting the last 'delay-1' bits */
	/* Current time information has been copied to the old time information, 
	so remember to use the old info.
	*/
	for(i=2;i<=Delay;i++)
	{
		DetectedHardOutput[SectorLength-1-(Delay-i)] = Path[0][MinPathMetricIndex][i-1];
		/* Detecting soft output */
		//if(DetectedHardOutput[SectorLength-1-(Delay-i)] == 1)
			DetectedSoftOutput[SectorLength-1-(Delay-i)] = ProbWrongDet[0][MinPathMetricIndex][i-1];
		//else
		//	DetectedSoftOutput[SectorLength-1-(Delay-i)] = 1 - ProbWrongDet[0][MinPathMetricIndex][i-1];
	}
	free(Path[0]);
	free(Path[1]);
	free(VS[0]);
	free(VS[1]);
	free(PathMetric[0]);
	free(PathMetric[1]);
	free(DifferingBits);
	for(i=0;i<NumStates;i++)
	{
		free(ProbWrongDet[0][i]);
		free(ProbWrongDet[1][i]);
	}
	free(ProbWrongDet[0]);
	free(ProbWrongDet[1]);
}


/*-------------------------------------- IS VALID STATE --------------------------------------*/

int IsValidState(int st,int mv,int *States,int *ModulationValues,int NumStates)
{
	int i;

	for(i=0;i<NumStates;i++)
	{
		if(st==States[i])
			if(mv == ModulationValues[i])
				return i;
	}
	return -1;	/* Reaches here if state is not valid */
}


/*---------------------------- VITERBI DETECTOR (DIFF IMPLEMENTATION) ------------------------*/

/* IMPORTANT: Use this version of viterbi if the Viterbi Delay is more than 20 */
/* IMPORTANT: Hard coded for state length 5 and for EEPR4 */

#define TEMPMOD_DEBUG 1

void ClassicalViterbiForTempMod(int Delay, double *EqualizedChannelOutput, int SectorLength, int *DetectedOutput, double *PRImpRes, int PRImpResLength, 
								int *States, int *ModulationValues, int **ValidIncomingTrans, int *NumberOfStates)
{
	int i,j,k,m,NumStates;
	double metric0,metric1,metric2;
	double sample0,sample1,sample2;
	double *PathMetric[2];
	int **Path[2];
	int MinPathMetricIndex;
	int StateLength = 9;								/* User input. Number of bits in a state */
	int IndexOfCentreBit = 6;							/* User input. Starting index is 1 */
	int TempModBitPosition = 4;							/* User input. Starting index is 0 */
	int temp,Index;
	int *NumValidIncomingTrans;
	static int ReadFileFlag = 0;
	char fname[]="ValidSequencesOfLength9.dat";			/* User input */
	double *EqualizerError;
	FILE *fp,*fpDebug;

	/* Hard Coding */
	EqualizerError = (double*)malloc((12)*sizeof(double));
	EqualizerError[0] = -0.0712;
	EqualizerError[1] = -0.072;
	EqualizerError[2] = -0.151;
	EqualizerError[3] = -0.453;
	EqualizerError[4] = -0.0979;
	EqualizerError[5] = 1.0878;
	EqualizerError[6] = 1.0984;
	EqualizerError[7] = -0.0979;
	EqualizerError[8] = -0.453;
	EqualizerError[9] = -0.151;
	EqualizerError[10] = -0.072;
	EqualizerError[11] = -0.0712;

	if(TEMPMOD_DEBUG)
	{
		fpDebug = fopen("Viterbi_TempMod_DebugFile.dat","w");
		if(fpDebug==NULL)
		{
			printf("ERROR: Cannot open file 'Viterbi_TempMod_DebugFile.dat'\n");
			getchar();
		}
	}

	NumStates = *NumberOfStates;

	/* Execute only once */
	if(ReadFileFlag==0)
	{
		ReadFileFlag=1;
		fp = fopen(fname,"r");
		if(fp==NULL)
		{
			printf("ERROR: Cannot open file %s\n",fname);
			getchar();
		}
//		States = (int*)malloc(sizeof(int));				/* Valid States */
//		ModulationValues = (int*)malloc(sizeof(int));	/* Indication of which bits are temperature modulated */
		NumStates=0;
		/* Reading all valid states from a file */
		while(!feof(fp))
		{
			temp = 0;
			NumStates++;
//			States = (int*)realloc(States,NumStates*sizeof(int));
//			ModulationValues = (int*)realloc(ModulationValues,NumStates*sizeof(int));
			States[NumStates-1]=0;
			ModulationValues[NumStates-1]=0;
			for(i=0;i<StateLength;i++)
			{
				fscanf(fp,"%1d",&temp);
				States[NumStates-1] = ((States[NumStates-1]) << 1)|temp;
			}
			for(i=0;i<StateLength;i++)
			{
				fscanf(fp,"%1d\n",&temp);
				ModulationValues[NumStates-1] = ((ModulationValues[NumStates-1]) << 1)|temp;
			}
		//	fscanf(fp,"%*[^\n]");
		}
		fclose(fp);
		*NumberOfStates = NumStates;
		
		/* Detecting all valid incoming transitions of all states */
//		ValidIncomingTrans = (int**)realloc(ValidIncomingTrans,NumStates*sizeof(int));
//		for(i=0;i<NumStates;i++)
//		{
//			ValidIncomingTrans[i] = (int*)malloc(3*sizeof(int));
//			ValidIncomingTrans[i][0]=-1;ValidIncomingTrans[i][1]=-1;ValidIncomingTrans[i][2]=-1;
//		}
		NumValidIncomingTrans = (int*)malloc(NumStates*sizeof(int));
		for(i=0;i<NumStates;i++)
		{
			NumValidIncomingTrans[i]=0;
			/* Transition bit 0 */
			Index = IsValidState((States[i]>>1),(ModulationValues[i]>>1),States,ModulationValues,NumStates);
			if(Index!=-1)	/* Valid transition */
			{
				NumValidIncomingTrans[i]++;
				ValidIncomingTrans[i][0]=Index;	/* Assign index of the state */
			}
			/* Transition bit 1 */
			Index = IsValidState((((States[i])>>1)|(1<<(StateLength-1))),((ModulationValues[i])>>1),States,ModulationValues,NumStates);
			if(Index!=-1)	/* Valid transition */
			{
				NumValidIncomingTrans[i]++;
				ValidIncomingTrans[i][1]=Index;	/* Assign index of the state */
			}
			/* Transition bit 2, i.e. corresponding temperature modulated bit */
			Index = IsValidState((((States[i])>>1)|((((States[i])>>(StateLength-1))&1)<<(StateLength-1))),(((ModulationValues[i])>>1)|(1<<(StateLength-1))),States,ModulationValues,NumStates);
			if(Index!=-1)	/* Valid transition */
			{
				NumValidIncomingTrans[i]++;
				ValidIncomingTrans[i][2]=Index;	/* Assign index of the state */
			}
		}
		
		if(TEMPMOD_DEBUG)
		{
			/* Prints states and the associated connections */
			for(i=0;i<NumStates;i++)
			{
				for(j=0;j<StateLength;j++)
					fprintf(fpDebug,"%d",(States[i]>>(StateLength-1-j))&1);
				fprintf(fpDebug," %d ",States[i]);
				for(j=0;j<StateLength;j++)
					fprintf(fpDebug,"%d",(ModulationValues[i]>>(StateLength-1-j))&1);
				fprintf(fpDebug," %d ",ModulationValues[i]);
				fprintf(fpDebug,">>>>>>>>> ");
				fprintf(fpDebug,"%d %d %d\n",ValidIncomingTrans[i][0],ValidIncomingTrans[i][1],ValidIncomingTrans[i][2]);
			}
			fprintf(fpDebug,"\n");
		}
	}

	/* 
	Path[0][i][j] : Bit at a delay (Delay-j) of the survivor path corresponding to state i
					at the previous time.
	Path[1][i][j] : Bit at a delay (Delay-j) of the survivor path corresponding to state i
					at the current time.
	*/
	Path[0] = (int**)malloc(NumStates*sizeof(int));
	Path[1] = (int**)malloc(NumStates*sizeof(int));
	for(i=0;i<NumStates;i++)
	{
		Path[0][i] = (int*)malloc(Delay*sizeof(int));
		Path[1][i] = (int*)malloc(Delay*sizeof(int));
	}
	for(i=0;i<NumStates;i++)
	{
		for(j=0;j<Delay;j++)
		{
			Path[0][i][j] = 0;
			Path[1][i][j] = 0;
		}
	}
	PathMetric[0] = (double *)malloc(NumStates * sizeof(double));
	PathMetric[1] = (double *)malloc(NumStates * sizeof(double));
	/* PathMetric[i] => The path metric of the valid output of state i at current time */
	/*
	NOTE that Pathmetric is initialized to 0 in the hope that by the time 'Delay' bits are
	transmitted, all of these are overwritten. If not, then when making a decision on a bit,
	the state for which Pathmetric is still 0 will be chosen, which is wrong. It cannot be
	initialized to a large number to avoid this problem because it is additive.
	*/
	for(i=0;i<NumStates;i++)
	{
		PathMetric[0][i] = 1e50;
		PathMetric[1][i] = 1e50;
	}
	PathMetric[0][0]=0;
	PathMetric[1][0]=0;
	
	/**** Main VITERBI loop ****/
	for(k=0;k<SectorLength;k++)
	{/* Loop over PaddedEqOutput */

		if(TEMPMOD_DEBUG)
			if(k%100==0 || k==2039)
				fprintf(fpDebug,"\n----------------------- Bit Number %d ----------------------\n\n",k);

		for(i=0;i<NumStates;i++)
		{/* Loop over all states */

			/*** For each state, calculating the path metric of the two incoming paths ***/

			/*
			If i = [i1 i2 ... in] (the array elements are binary) is the current state
			then its corresponding previous states are [0 i1 i2.....in-1] and [1 i1 i2....in-1].
			These are determined by	bit operations given below.
			*/

			/**
			Calculating the sample value and metric at current time for the transition from
			state [0 i1 i2....in-1]	to state [i1 i2....in].
			Corresponding Previous state is given by: (CurrentState >> 1) => shifting all
			bits to the right by 1 position and appending a ZERO at the left most position.
			**/
			if(ValidIncomingTrans[i][0] != -1)	/* If it is a valid transition */
			{
				sample0 = 0.0;
				for(j=0;j<=PRImpResLength-1;j++)
				{
					sample0 += PRImpRes[j]*((States[i] >> (j+StateLength-IndexOfCentreBit)) & 1);
				}
				sample0 = sample0*2;
				/* Account for any temperature modulated bits in the state */
				if(ModulationValues[i]>0)
				{
					for(j=0;j<StateLength;j++)
					{
						if((ModulationValues[i] >> j)&1 == 1)	/* Temp mod bit */
							sample0 = sample0 - (((States[i]>>j)&1) - !((States[i]>>j)&1))*EqualizerError[TempModBitPosition+IndexOfCentreBit-(StateLength-j)];
					}
				}
				metric0 = PathMetric[0][ValidIncomingTrans[i][0]] + pow((EqualizedChannelOutput[k] - sample0),2);
			}else
			{
				metric0 = 1e50;
			}

			/**
			Calculating the sample value and metric at current time for the transition from
			state [1 i1 i2....in-1]	to state [i1 i2....in]
			Corresponding Previous state is given by: (CurrentState >> 1) | (1 << (CurrentStateLength-1))
			=> shifting all bits to the right by 1 position and appending a ONE at the
			left most position
			**/
			if(ValidIncomingTrans[i][1] != -1)	/* If it is a valid transition */
			{
				sample1 = 0.0;
				for(j=0;j<=PRImpResLength-1;j++)
				{
					sample1 += PRImpRes[j]*((States[i] >> (j+StateLength-IndexOfCentreBit)) & 1);
				}
				sample1 = sample1*2;
				/* Account for any temperature modulated bits in the state */
				if(ModulationValues[i]>0)
				{
					for(j=0;j<StateLength;j++)
					{
						if((ModulationValues[i] >> j)&1 == 1)	/* Temp mod bit */
							sample1 = sample1 - (((States[i]>>j)&1) - !((States[i]>>j)&1))*EqualizerError[TempModBitPosition+IndexOfCentreBit-(StateLength-j)];
					}
				}
				metric1 = PathMetric[0][ValidIncomingTrans[i][1]] + pow((EqualizedChannelOutput[k] - sample1),2);
			}else
			{
				metric1 = 1e50;
			}

			/**
			Calculating the sample value and metric at current time for the transition from
			state [2 i1 i2....in-1]	to state [i1 i2....in]. '2' is the corresponding temperature modulated
			bit.
			**/
			if(ValidIncomingTrans[i][2] != -1)	/* If it is a valid transition */
			{
				sample2 = 0.0;
				for(j=0;j<=PRImpResLength-1;j++)
				{
					sample2 += PRImpRes[j]*((States[i] >> (j+StateLength-IndexOfCentreBit)) & 1);
				}
				sample2 = sample2*2;
				/* Account for any temperature modulated bits in the state */
				if(ModulationValues[i]>0)
				{
					for(j=0;j<StateLength;j++)
					{
						if((ModulationValues[i] >> j)&1 == 1)	/* Temp mod bit */
							sample2 = sample2 - (((States[i]>>j)&1) - !((States[i]>>j)&1))*EqualizerError[TempModBitPosition+IndexOfCentreBit-(StateLength-j)];
					}
				}
				sample2 = sample2 - ((((States[i]) >> (StateLength-1))&1) - !((States[i]) >> (StateLength-1))&1)*EqualizerError[TempModBitPosition+IndexOfCentreBit];
				metric2 = PathMetric[0][ValidIncomingTrans[i][2]] + pow((EqualizedChannelOutput[k] - sample2),2);
			}else
			{
				metric2 = 1e50;
			}

			if(metric0 < 1e50 || metric1 < 1e50 || metric2 < 1e50)	/* Execute atleast when one of the incoming transition is valid */
			{
				if(TEMPMOD_DEBUG)
				{
					if(k%100==0 || k==2039)
						fprintf(fpDebug,"%d %d %g %g %g %g %g %g %g ",States[i],ModulationValues[i],EqualizedChannelOutput[k],sample0,sample1,sample2,metric0,metric1,metric2);
				}

				if((metric0 <= metric1) && (metric0 <= metric2))
				{
					/*
					New path is from the state [0 i1 i2...in-1], i.e current valid path for state i is
					the valid path of state [0 i1...in-1] in the previous time appended with bit 'in'
					*/
					/*
					(Path[0][i >> 1] << 1) => shifting all bits of path of the state [0 i1...in-1]
					to the left by 1 position.
					(i & 1) => The last bit of state i.
					(Path[0][i >> 1] << 1) | (i & 1) => appending last bit of i to the left-shifted path.
					*/
					for(m=0;m<Delay-1;m++)
					{
						Path[1][i][m] = Path[0][ValidIncomingTrans[i][0]][m+1];
					}
					Path[1][i][Delay-1] = (ModulationValues[i] & 1)==1 ? 2:(States[i] & 1);
					PathMetric[1][i] = metric0;
				}else if((metric1 <= metric0) && (metric1 <= metric2))
				{
					/*
					New path is from the state [1 i1 i2...in-1], i.e current valid path for state i is
					the valid path of state [1 i1...in-1] in the previous time appended with bit 'in'
					*/
					/* Bit operations are done as before */
					for(m=0;m<Delay-1;m++)
					{
						Path[1][i][m] = Path[0][ValidIncomingTrans[i][1]][m+1];
					}
					Path[1][i][Delay-1] = (ModulationValues[i] & 1)==1 ? 2:(States[i] & 1);
					PathMetric[1][i] = metric1;
				}else
				{
					for(m=0;m<Delay-1;m++)
					{
						Path[1][i][m] = Path[0][ValidIncomingTrans[i][2]][m+1];
					}
					Path[1][i][Delay-1] = (ModulationValues[i] & 1)==1 ? 2:(States[i] & 1);
					PathMetric[1][i] = metric2;
				}
				if(TEMPMOD_DEBUG)
				{
					if(k%100==0 || k==2039)
						fprintf(fpDebug," %g\n",PathMetric[1][i]);
				}
			}

		}/* Loop over all states */

		/* Make a decision on the bit transmitted 'Delay' bit periods before */
		if(k >= Delay-1)
		{
			/** Deciding on the transmitted bit based on detection delay **/
			/* Finding the minimum path metric at the current time */
			MinPathMetricIndex = 0;
			for(i=1;i<NumStates;i++)
			{
				if(PathMetric[1][i] < PathMetric[1][MinPathMetricIndex])
					MinPathMetricIndex = i;
			}
//			printf("Running temp code: Printing Minimum path metric\n");
//			printf("Minimum path metric = %g\n",PathMetric[1][MinPathMetricIndex]);
			/* Detected output = first bit in Path[1][MinPathMetricIndex] */
			/* Path[1][MinPathMetricIndex] has 'Delay' number of bits */
			DetectedOutput[k-(Delay-1)] = Path[1][MinPathMetricIndex][0];
		}

		/* Exchange old and new Path and PathMetric values */
		for(i=0;i<NumStates;i++)
		{
			for(m=0;m<Delay;m++)
			{
				Path[0][i][m] = Path[1][i][m];
				Path[1][i][m] = 0;
			}
			PathMetric[0][i] = PathMetric[1][i];
			PathMetric[1][i] = 0.0;
		}
		
	}/* Loop over PaddedEqOutput */

	/* Detecting the last 'delay-1' bits */
	for(i=2;i<=Delay;i++)
		DetectedOutput[SectorLength-1-(Delay-i)] = Path[0][MinPathMetricIndex][i-1];

	/* Account for delay due to causality */
	/* IMPORTANT: Assuming zero pre and post padding */
	for(i=SectorLength-1;i>=(StateLength-IndexOfCentreBit);i--)
		DetectedOutput[i]=DetectedOutput[i-(StateLength-IndexOfCentreBit)];
	for(i=0;i<(StateLength-IndexOfCentreBit);i++)
		DetectedOutput[i]=0;		/* Assuming zero padding */

	if(TEMPMOD_DEBUG)
	{
		for(i=0;i<NumStates;i++)
			if(PathMetric[0][i]==PathMetric[0][MinPathMetricIndex])
			{
				fprintf(fpDebug,"%d ",States[i]);
				for(j=0;j<Delay;j++)
					fprintf(fpDebug,"%d ", Path[0][MinPathMetricIndex][j]);
				fprintf(fpDebug,"\n");
			}
	}

	/* Converting temp mod bits to tribits */
	for(i=1;i<SectorLength;i++)
	{
		if(DetectedOutput[i] == 2)
		{
			DetectedOutput[i] = DetectedOutput[i-1];
			DetectedOutput[i-1] = abs(DetectedOutput[i]-1);
		}
	}


	for(i=0;i<NumStates;i++)
	{
		free(Path[0][i]);
		free(Path[1][i]);
	}

	if(TEMPMOD_DEBUG)
		fclose(fpDebug);

	free(Path[0]);
	free(Path[1]);
	free(PathMetric[0]);
	free(PathMetric[1]);
	free(NumValidIncomingTrans);
	free(EqualizerError);
}

/*---------------------------- VITERBI DETECTOR (DIFF IMPLEMENTATION) ------------------------*/

/* IMPORTANT: Use this version of viterbi if the Viterbi Delay is more than 20 */

void ClassicalViterbiForTempModWithPerm(int Delay, double *EqualizedChannelOutput, int SectorLength, int *DetectedOutput, double *PRImpRes, 
										int PRImpResLength, double *MSE, int NumBitsRepPermPeriod, int SubSectorLength, int NumInterleavedBits, int PrePaddingLength)
{
	int i,j,k,m,NumStates,CurrentSubSector;
	double metric0,metric1;
	double sample0,sample1;
	double *PathMetric[2];
	int **Path[2];
	int MinPathMetricIndex;

	/* Number of bits in every state = PRImpResLength-1 (= Order of PR polynomial) */
	/* Number of states in trellis = 2^(Order of PR polynomial) */
	NumStates = (int)pow(2,(PRImpResLength-1));
	Delay = SubSectorLength;

	/* 
	Path[0][i][j] : Bit at a delay (Delay-j) of the survivor path corresponding to state i
					at the previous time.
	Path[1][i][j] : Bit at a delay (Delay-j) of the survivor path corresponding to state i
					at the current time.
	*/
	Path[0] = (int**)malloc(NumStates*sizeof(int));
	Path[1] = (int**)malloc(NumStates*sizeof(int));
	for(i=0;i<NumStates;i++)
	{
		Path[0][i] = (int*)malloc(Delay*sizeof(int));
		Path[1][i] = (int*)malloc(Delay*sizeof(int));
	}
	for(i=0;i<NumStates;i++)
	{
		for(j=0;j<Delay;j++)
		{
			Path[0][i][j] = 0;
			Path[1][i][j] = 0;
		}
	}
	PathMetric[0] = (double *)malloc(NumStates * sizeof(double));
	PathMetric[1] = (double *)malloc(NumStates * sizeof(double));
	/* PathMetric[i] => The path metric of the valid output of state i at current time */
	/*
	NOTE that Pathmetric is initialized to 0 in the hope that by the time 'Delay' bits are
	transmitted, all of these are overwritten. If not, then when making a decision on a bit,
	the state for which Pathmetric is still 0 will be chosen, which is wrong. It cannot be
	initialized to a large number to avoid this problem because it is additive.
	*/
	for(i=0;i<NumStates;i++)
	{
		PathMetric[0][i] = 1e50;
		PathMetric[1][i] = 1e50;
	}
	PathMetric[0][0]=0;
	PathMetric[1][0]=0;
	
	CurrentSubSector = 1;

	/**** Main VITERBI loop ****/
	for(k=0;k<SectorLength;k++)
	{/* Loop over PaddedEqOutput */
		
		for(i=0;i<NumStates;i++)
		{/* Loop over all states */

			/*** For each state, calculating the path metric of the two incoming paths ***/

			/*
			If i = [i1 i2 ... in] (the array elements are binary) is the current state
			then its corresponding previous states are [0 i1 i2.....in-1] and [1 i1 i2....in-1].
			These are determined by	bit operations given below.
			*/

			/**
			Calculating the sample value and metric at current time for the transition from
			state [0 i1 i2....in-1]	to state [i1 i2....in].
			Corresponding Previous state is given by: (CurrentState >> 1) => shifting all
			bits to the right by 1 position and appending a ZERO at the left most position.
			**/
			sample0 = 0.0;
			for(j=0;j<PRImpResLength-1;j++)
			{
				sample0 += PRImpRes[j]*((i >> j) & 1);
			}
			sample0 += PRImpRes[PRImpResLength-1]*0;	/* Eventhough it is 0, it is explicitly written for clarity */
			sample0 = sample0*2;
			metric0 = PathMetric[0][i >> 1] + pow((EqualizedChannelOutput[k] - sample0),2);

			/**
			Calculating the sample value and metric at current time for the transition from
			state [1 i1 i2....in-1]	to state [i1 i2....in]
			Corresponding Previous state is given by: (CurrentState >> 1) | (1 << (CurrentStateLength-1))
			=> shifting all bits to the right by 1 position and appending a ONE at the
			left most position
			**/
			sample1 = 0.0;
			for(j=0;j<PRImpResLength-1;j++)
			{
				sample1 += PRImpRes[j]*((i >> j) & 1);
			}
			sample1 += PRImpRes[PRImpResLength-1]*1;
			sample1 = sample1*2;
			metric1 = PathMetric[0][(i >> 1)|(1 << (PRImpResLength-2))] + pow((EqualizedChannelOutput[k] - sample1),2);

			/* Select the path with smaller metric among the two incoming paths */
			if(metric0 <= metric1)
			{
				/*
				New path is from the state [0 i1 i2...in-1], i.e current valid path for state i is
				the valid path of state [0 i1...in-1] in the previous time appended with bit 'in'
				*/
				/*
				(Path[0][i >> 1] << 1) => shifting all bits of path of the state [0 i1...in-1]
				to the left by 1 position.
				(i & 1) => The last bit of state i.
				(Path[0][i >> 1] << 1) | (i & 1) => appending last bit of i to the left-shifted path.
				*/
				for(m=0;m<Delay-1;m++)
				{
					Path[1][i][m] = Path[0][i >> 1][m+1];
				}
				Path[1][i][Delay-1] = i & 1;
				PathMetric[1][i] = metric0;
			}else{
				/*
				New path is from the state [1 i1 i2...in-1], i.e current valid path for state i is
				the valid path of state [1 i1...in-1] in the previous time appended with bit 'in'
				*/
				/* Bit operations are done as before */
				for(m=0;m<Delay-1;m++)
				{
					Path[1][i][m] = Path[0][(i >> 1)|(1 << (PRImpResLength-2))][m+1];
				}
				Path[1][i][Delay-1] = i & 1;
				PathMetric[1][i] = metric1;
			}
		}/* Loop over all states */

		/* If last bit of the current subsector is passed through the trellis (excluding the interleaved bits in this sector) */
		if(k-PrePaddingLength == SubSectorLength + (SubSectorLength+NumInterleavedBits)*(CurrentSubSector-1) - 1 && CurrentSubSector!=NumBitsRepPermPeriod)
		{
			/* Make a decision on the current subsector bits */
			/* Finding the minimum path metric at the current time */
			MinPathMetricIndex = 0;
			for(i=1;i<NumStates;i++)
			{
				if(PathMetric[1][i] < PathMetric[1][MinPathMetricIndex])
					MinPathMetricIndex = i;
			}

			for(i=0;i<Delay;i++)
			{
				DetectedOutput[k-(Delay-1-i)] = Path[1][MinPathMetricIndex][i];
			}
			
			/* Store the minimum MSE for the current sub-sector */
			MSE[CurrentSubSector-1] = PathMetric[1][MinPathMetricIndex];

			/* Next 'NumInterleavedBits' bits are 0's */
			k = k + NumInterleavedBits;	

			/* Re-initialize path metric */
			for(i=0;i<NumStates;i++)
			{
				PathMetric[0][i] = 1e50;
				PathMetric[1][i] = 1e50;
			}
			PathMetric[0][0]=0;
			PathMetric[1][0]=0;
			/* Re-initialize path */
			for(i=0;i<NumStates;i++)
			{
				for(j=0;j<Delay;j++)
				{
					Path[0][i][j] = 0;
					Path[1][i][j] = 0;
				}
			}
			CurrentSubSector++;
		}		
		
		/* Condition check for the last interleaved bit of the last sub-sector */
		if((k-PrePaddingLength == (SubSectorLength+NumInterleavedBits)*(CurrentSubSector-1)-1) && (CurrentSubSector==NumBitsRepPermPeriod))
		{
			/* Last sub-sector is not of length SubSectorLength */
			Delay = SectorLength - k - 1;	/* Length of the last sub-sector */
			for(i=0;i<NumStates;i++)
			{
				Path[0][i] = (int*)realloc(Path[0][i],Delay*sizeof(int));
				Path[1][i] = (int*)realloc(Path[1][i],Delay*sizeof(int));
			}
			for(i=0;i<NumStates;i++)
			{
				for(j=0;j<Delay;j++)
				{
					Path[0][i][j] = 0;
					Path[1][i][j] = 0;
				}
			}
		}

		/* Detect the last sub-sector */
		if(k==SectorLength-1)
		{
			MinPathMetricIndex = 0;
			for(i=1;i<NumStates;i++)
			{
				if(PathMetric[1][i] < PathMetric[1][MinPathMetricIndex])
					MinPathMetricIndex = i;
			}
			for(i=0;i<Delay;i++)
			{
				DetectedOutput[k-(Delay-1-i)] = Path[1][MinPathMetricIndex][i];
			}			
			/* Store the minimum MSE for the current sub-sector */
			MSE[CurrentSubSector-1] = PathMetric[1][MinPathMetricIndex];
		}

		/* Exchange old and new Path and PathMetric values */
		for(i=0;i<NumStates;i++)
		{
			for(m=0;m<Delay;m++)
			{
				Path[0][i][m] = Path[1][i][m];
				Path[1][i][m] = 0;
			}
			PathMetric[0][i] = PathMetric[1][i];
			PathMetric[1][i] = 0.0;
		}

	}/* Loop over PaddedEqOutput */

	for(i=0;i<NumStates;i++)
	{
		free(Path[0][i]);
		free(Path[1][i]);
	}
	free(Path[0]);
	free(Path[1]);
	free(PathMetric[0]);
	free(PathMetric[1]);
}