#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include "MagneticDisk.h"

/*------------------------- DECODER FOR 4/5 RLL (0,2) CODE ------------------------------*/

int Dec_4By5RLLCode(int *NRZPaddedViterbiOutput, int *NRZDecodedOutput, int PrePaddingLength, int SectorLength)
{
	int i,j,k,dec,ValidCodeword;
	char temp[5],dummy;
	int *NRZIDecodedOutput,*NRZViterbiOutput, *NRZIViterbiOutput;
	static int Codewords[16][6], ReadFlag=0;	/* Codeword[i][6] stores the cooresponding decimal value of the codeword */
	double rate = 4.0/5.0;
	int InvalidCW = 0;	
	char *RLLCodeFileName = "./Codes/Rate4By5RLLCode.dat";
	FILE *fCodewords = NULL;


	if(ReadFlag==0)
	{
		ReadFlag=1;
		fCodewords = fopen(RLLCodeFileName,"r");
		if(fCodewords == NULL)
		{
			printf("ERROR: -> Dec_4By5RLLCode() -> Cannot open file %s \n",RLLCodeFileName);
			exit(-1);
		}
		
		/* Reading codewords from the file */
		for(i=0;i<16;i++)
		{
			fscanf(fCodewords,"%s\n",temp);
			for(j=0;j<5;j++)
			{
				dummy = temp[j];
				Codewords[i][j] = atoi(&(dummy));
			}		
			Codewords[i][5] = 0;
			for(j=0;j<5;j++)
				Codewords[i][5] += (int)pow(2,4-j)*Codewords[i][j];
		}
		fclose(fCodewords);
	}

	NRZIViterbiOutput = (int*)malloc((SectorLength-1) * sizeof(int));
	NRZViterbiOutput = (int*)malloc(SectorLength * sizeof(int));
	NRZIDecodedOutput = (int*)malloc((int)((SectorLength-1)*rate) * sizeof(int));

	for(i=PrePaddingLength;i<SectorLength+PrePaddingLength;i++)
		NRZViterbiOutput[i-PrePaddingLength] = NRZPaddedViterbiOutput[i];

	/* Convert NRZ viterbi output to NRZI format */
	for(i=1;i<SectorLength;i++)
	{
		NRZIViterbiOutput[i-1] = abs(NRZViterbiOutput[i]-NRZViterbiOutput[i-1]);
	}
	
	/* Decoding */
	for(i=0;i<(SectorLength-1)/5;i++)
	{
		dec = 0;
		/* Decimal value of the 5-bit block */
		for(j=0;j<5;j++)
			dec += (int) pow(2,4-j)*NRZIViterbiOutput[i*5+j];
		/* Searching for the corresponding 6-bit user word */
		ValidCodeword = 0;
		for(j=0;j<16;j++)
		{
			if(dec == Codewords[j][5])	
			{
				ValidCodeword = 1;
				/* Sent 4-bit user word has the decimal value j */
				for(k=0;k<4;k++)
					NRZIDecodedOutput[i*4+k] = (j >> (3-k)) & 1;	/* Extracting all 4-bits from j */
				break;
			}
		}
		if(ValidCodeword == 0)
		{
			printf("WARNING -> Dec_4By5RLLCode() -> InValid codeword encountered.\n");
			printf("Decimal value of the codeword is %d \n",dec);
			/* Just send all-zeros */
			InvalidCW = 1;
			for(k=0;k<4;k++)
				NRZIDecodedOutput[i*4+k] = 0;
		}
	}

	/* Converting NRZI decoded bits to NRZ decoded bits */
	NRZDecodedOutput[0] = 0;
	for(i=0;i<(int)((SectorLength-1)*rate);i++)
	{
		NRZDecodedOutput[i+1] = (NRZIDecodedOutput[i] == 0) ? NRZDecodedOutput[i] : (NRZDecodedOutput[i]+1)%2;
	}
	free(NRZIDecodedOutput);
	free(NRZViterbiOutput);
	free(NRZIViterbiOutput);
	return(InvalidCW);
}


/*------------------------- DECODER FOR 6/7 MTR(2;8) CODE ------------------------------*/
/* 
This function implements the decoder for MTR(2;8) rate 6/7 code.
For more info see: Barrett Brickner, Jaekyun Moon, "Design of rate 6/7 Max Trans Run Code", 
IEEE Trans Mag, Vol 33, No 5, Sept 1997.
*/
void Dec_6By7MTRCode(int *NRZPaddedViterbiOutput, int *NRZDecodedOutput, int PrePaddingLength, int SectorLength)
{
	int i,j,k,dec,ValidCodeword;
	char temp[7],dummy;
	int *NRZIDecodedOutput,*NRZViterbiOutput, *NRZIViterbiOutput;
	int Codewords[64][8];	/* Codeword[i][7] stores the cooresponding decimal value of the codeword */
	double rate = 6.0/7.0;
	FILE *fCodewords = NULL;
	
	fCodewords = fopen("Codes/Rate6By7MTR2-8.dat","r");
	if(fCodewords == NULL)
	{
		printf("ERROR: -> Enc_6By7MTR2-8Code() -> Cannot open file %s \n","Codes/Rate6By7MTR2-8.dat");
		exit(-1);
	}
	
	/* Reading codewords from the file */
	for(i=0;i<64;i++)
	{
		fscanf(fCodewords,"%s\n",temp);
		for(j=0;j<7;j++)
		{
			dummy = temp[j];
			Codewords[i][j] = atoi(&(dummy));
		}		
		Codewords[i][7] = 0;
		for(j=0;j<7;j++)
			Codewords[i][7] += (int)pow(2,6-j)*Codewords[i][j];
	}
	fclose(fCodewords);

	NRZIViterbiOutput = (int*)malloc((SectorLength-1) * sizeof(int));
	NRZViterbiOutput = (int*)malloc(SectorLength * sizeof(int));
	NRZIDecodedOutput = (int*)malloc((int)((SectorLength-1)*rate) * sizeof(int));

	for(i=PrePaddingLength;i<SectorLength+PrePaddingLength;i++)
		NRZViterbiOutput[i-PrePaddingLength] = NRZPaddedViterbiOutput[i];

	/* Convert NRZ viterbi output to NRZI format */
	for(i=1;i<SectorLength;i++)
	{
		NRZIViterbiOutput[i-1] = abs(NRZViterbiOutput[i]-NRZViterbiOutput[i-1]);
	}
	
	/* Detect and undo substitutions */
	for(i=0;i<(SectorLength-1)/7 - 1;i++)
	{
		/* Detecting substitutions */
		if((NRZIViterbiOutput[i*7+4]==0) && (NRZIViterbiOutput[i*7+5]==1) && (NRZIViterbiOutput[i*7+6]==1))      
		{
			/* Detecting and undoing substitution Type I */
			if((NRZIViterbiOutput[(i+1)*7]==0) && (NRZIViterbiOutput[(i+1)*7+1]==0) && (NRZIViterbiOutput[(i+1)*7+2]==1))
			{
				NRZIViterbiOutput[i*7+4] = 0;
				NRZIViterbiOutput[i*7+5] = 0;
				NRZIViterbiOutput[i*7+6] = 1;
				NRZIViterbiOutput[(i+1)*7] = 1;
				NRZIViterbiOutput[(i+1)*7+1] = 1;
				NRZIViterbiOutput[(i+1)*7+2] = 0;
				continue;
			}
			/* Detecting and undoing substitution Type II */
			if((NRZIViterbiOutput[(i+1)*7]==0) && (NRZIViterbiOutput[(i+1)*7+1]==1) && (NRZIViterbiOutput[(i+1)*7+2]==0))
			{
				NRZIViterbiOutput[i*7+4] = 1;
				NRZIViterbiOutput[i*7+5] = 0;
				NRZIViterbiOutput[i*7+6] = 1;
				NRZIViterbiOutput[(i+1)*7] = 1;
				NRZIViterbiOutput[(i+1)*7+1] = 1;
				NRZIViterbiOutput[(i+1)*7+2] = 0;
				continue;
			}
			/* Detecting and undoing substitution Type III */
			if((NRZIViterbiOutput[(i+1)*7]==0) && (NRZIViterbiOutput[(i+1)*7+1]==0) && (NRZIViterbiOutput[(i+1)*7+2]==0))
			{
				NRZIViterbiOutput[i*7+4] = 0;
				NRZIViterbiOutput[i*7+5] = 0;
				NRZIViterbiOutput[i*7+6] = 0;
				NRZIViterbiOutput[(i+1)*7] = 0;
				NRZIViterbiOutput[(i+1)*7+1] = 0;
				NRZIViterbiOutput[(i+1)*7+2] = 0;
				continue;
			}
		}
	}

	/* Decoding */
	for(i=0;i<(SectorLength-1)/7;i++)
	{
		dec = 0;
		/* Decimal value of the 7-bit block */
		for(j=0;j<7;j++)
			dec += (int) pow(2,6-j)*NRZIViterbiOutput[i*7+j];
		/* Searching for the corresponding 6-bit user word */
		ValidCodeword = 0;
		for(j=0;j<64;j++)
		{
			if(dec == Codewords[j][7])	
			{
				ValidCodeword = 1;
				/* Sent 6-bit user word has the decimal value j */
				for(k=0;k<6;k++)
					NRZIDecodedOutput[i*6+k] = (j >> (5-k)) & 1;	/* Extracting all 6-bits from j */
				break;
			}
		}
		if(ValidCodeword == 0)
		{
			printf("ERROR -> Dec_6By7MTRCode() -> InValid codeword encountered. Terminating...\n");
			printf("Decimal value of the codeword is %d \n",dec);
			//exit(-1);
			for(k=0;k<7;k++)
				printf("%d",NRZIViterbiOutput[i*7+k]);
			printf("\n");
			//exit(-1);
			/* Just send all-zeros */
			for(k=0;k<6;k++)
				NRZIDecodedOutput[i*6+k] = 0;
		}
	}

	/* Converting NRZI decoded bits to NRZ decoded bits */
	NRZDecodedOutput[0] = 0;
	for(i=0;i<(int)((SectorLength-1)*rate);i++)
	{
		NRZDecodedOutput[i+1] = (NRZIDecodedOutput[i] == 0) ? NRZDecodedOutput[i] : (NRZDecodedOutput[i]+1)%2;
	}
	free(NRZIDecodedOutput);
	free(NRZViterbiOutput);
	free(NRZIViterbiOutput);
	return;
}


/*------------------------- DECODER FOR 8/9 TMTR(2/3;11) CODE ------------------------------*/
/* 
This function implements the decoder for TMTR(2/3;11) rate 8/9 code.
For more info see: William Bliss, "An 8/9 rate time varying rellis code for
high density magnetic recording", IEEE Trans Mag, Vol 33, No 5, Sept 1997.
*/
int Dec_8By9TMTRCode(int *NRZPaddedViterbiOutput, int *NRZDecodedOutput, int PrePaddingLength, int SectorLength)
{
	int i,j,k,dec,ValidCodeword;
	char temp[9],dummy;
	int *NRZIDecodedOutput,*NRZViterbiOutput, *NRZIViterbiOutput;
	int Codewords[256][10];	/* Codeword[i][7] stores the cooresponding decimal value of the codeword */
	int InvalidCW = 0;
	double rate = 8.0/9.0;

	FILE *fCodewords = NULL;
	
	fCodewords = fopen("Codes/Rate8By9TMTR.dat","r");
	if(fCodewords == NULL)
	{
		printf("ERROR: -> Enc_8By9TMTR() -> Cannot open file %s \n","Codes/Rate8By9TMTR.dat");
		exit(-1);
	}
	
	/* Reading codewords from the file */
	for(i=0;i<256;i++)
	{
		fscanf(fCodewords,"%s\n",temp);
		for(j=0;j<9;j++)
		{
			dummy = temp[j];
			Codewords[i][j] = atoi(&(dummy));
		}		
		Codewords[i][9] = 0;
		/* Decimal value of the 9-bit word */
		for(j=0;j<9;j++)
			Codewords[i][9] += (int)pow(2,8-j)*Codewords[i][j];
	}
	fclose(fCodewords);

	NRZIViterbiOutput = (int*)malloc((SectorLength-1) * sizeof(int));
	NRZViterbiOutput = (int*)malloc(SectorLength * sizeof(int));
	NRZIDecodedOutput = (int*)malloc((int)((SectorLength-1)*rate) * sizeof(int));

	for(i=PrePaddingLength;i<SectorLength+PrePaddingLength;i++)
		NRZViterbiOutput[i-PrePaddingLength] = NRZPaddedViterbiOutput[i];

	/* Convert NRZ viterbi output to NRZI format */
	for(i=1;i<SectorLength;i++)
	{
		NRZIViterbiOutput[i-1] = abs(NRZViterbiOutput[i]-NRZViterbiOutput[i-1]);
	}
	
	/* Decoding */
	for(i=0;i<(SectorLength-1)/9;i++)
	{
		dec = 0;
		/* Decimal value of the 9-bit block */
		for(j=0;j<9;j++)
			dec += (int) pow(2,8-j)*NRZIViterbiOutput[i*9+j];
		
		ValidCodeword = 0;
		/* Searching for the corresponding 8-bit user word */
		for(j=0;j<256;j++)
		{
			if(dec == Codewords[j][9])	
			{
				ValidCodeword = 1;
				/* Sent 8-bit user word has the decimal value j */
				for(k=0;k<8;k++)
					NRZIDecodedOutput[i*8+k] = (j >> (7-k)) & 1;	/* Extracting all 6-bits from j */
				break;
			}
		}
		if(ValidCodeword == 0)
		{
			printf("ERROR: Dec_8By9TMTRCode() -> Invalid codeword encountered\n");
			printf("Decimal value of the codeword = %d\n",dec);
			for(k=0;k<9;k++)
				printf("%d",NRZIViterbiOutput[i*9+k]);
			printf("\n");
			//exit(-1);
			InvalidCW = 1;
			/* Just send all-zeros */
			for(k=0;k<8;k++)
				NRZIDecodedOutput[i*8+k] = 0;
		}
	}

	/* Converting NRZI decoded bits to NRZ decoded bits */
	NRZDecodedOutput[0] = 0;
	for(i=0;i<(int)((SectorLength-1)*rate);i++)
	{
		NRZDecodedOutput[i+1] = (NRZIDecodedOutput[i] == 0) ? NRZDecodedOutput[i] : (NRZDecodedOutput[i]+1)%2;
	}
	free(NRZIDecodedOutput);
	free(NRZViterbiOutput);
	free(NRZIViterbiOutput);
	return(InvalidCW);
}