#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include "MagneticDisk.h"

/*------------------------- ENCODER FOR 4/5 RLL(0;2) CODE ------------------------------*/

void Enc_4By5RLLCode(int *UserBits, int *EncodedBits, int *UserSecLen)
{
	int i,j,dec;
	static int UserSectorLength;
	char temp[5],dummy;
	int *NRZIUserBits,*NRZIEncodedBits;
	static int Codewords[16][5], ReadFlag=0;
	double rate = 4.0/5.0;
	char *RLLCodeFileName = "./Codes/Rate4By5RLLCode.dat";
	FILE *fCodewords = NULL;
	
	/* 
	NRZ User Sector Length (*UserSecLen) should be of the form 4Z+1, where Z is an integer.
	Because, NRZI sector length would then be 4Z. NRZI Encoded length will be 5Z
	and NRZ encoded length will be 5Z+1.
	*/

	if(ReadFlag==0)
	{
		ReadFlag=1;
		if((*UserSecLen) % 4 != 1)
		{
			printf("WARNING: -> Enc_4By5RLLCode() -> For 4/5 RLL(0,2) code, sector length should be 4Z+1, where Z is an integer \n");
			*UserSecLen = *UserSecLen - ((*UserSecLen)%4) + 1;
			printf("WARNING: -> Enc_4By5RLLCode() -> Sector length changed to %d \n",(*UserSecLen));
		}
		UserSectorLength = *UserSecLen;

		fCodewords = fopen(RLLCodeFileName,"r");
		if(fCodewords == NULL)
		{
			printf("ERROR: -> Enc_4By5RLLCode() -> Cannot open file %s \n",RLLCodeFileName);
			exit(-1);
		}
		
		/* Reading codewords from the file */
		for(i=0;i<16;i++)
		{
			fscanf(fCodewords,"%s\n",temp);
			for(j=0;j<5;j++)
			{
				dummy = temp[j];
				Codewords[i][j] = atoi(&(dummy));
			}
		}
		fclose(fCodewords);
	}

	NRZIUserBits = (int *) malloc((UserSectorLength-1) * sizeof(int));
	NRZIEncodedBits = (int *) malloc((int)((UserSectorLength-1)/rate) * sizeof(int));

	/* Convert NRZ userbits to NRZI userbits */
	for(i=1;i<UserSectorLength;i++)
	{
		NRZIUserBits[i-1]= abs(UserBits[i]-UserBits[i-1]);
	}

	/* Encoding the NRZI user bits */
	for(i=0;i<(UserSectorLength-1)/4;i++)
	{
		/* Decimal value of the 4-bit word */
		dec = 0;
		for(j=0;j<4;j++)
			dec = dec + (int) pow(2,3-j)*NRZIUserBits[i*4 + j];
		for(j=0;j<5;j++)
			NRZIEncodedBits[i*5 + j] = Codewords[dec][j];
	}

	/* Convert NRZI Encoded Bits to NRZ format */
	/* Assume that the first NRZ encoded bit is 0 */
	EncodedBits[0] = 0;
	for(i=0;i<(int)((UserSectorLength-1)/rate);i++)
	{
		EncodedBits[i+1] = (NRZIEncodedBits[i] == 0) ? EncodedBits[i] : (EncodedBits[i]+1)%2;
	}
	free(NRZIUserBits);
	free(NRZIEncodedBits);
	return;
}


/*------------------------- ENCODER FOR 6/7 MTR(2;8) CODE ------------------------------*/
/* 
This function implments the encoder for MTR(2;8) rate 6/7 code.
For more info see: Barrett Brickner, Jaekyun Moon, "Design of rate 6/7 Max Trans Run Code", 
IEEE Trans Mag, Vol 33, No 5, Sept 1997.
*/
void Enc_6By7MTRCode(int *UserBits, int *EncodedBits, int *UserSecLen)
{
	int i,j,dec, UserSectorLength;
	char temp[7],dummy;
	int *NRZIUserBits,*NRZIEncodedBits;
	int Codewords[64][7];
	double rate = 6.0/7.0;
	FILE *fCodewords = NULL;
	
	/* 
	NRZ User Sector Length (*UserSecLen) should be of the form 6Z+1, where Z is an integer.
	Because, NRZI sector length would then be 6Z. NRZI Encoded length will be 7Z
	and NRZ encoded length will be 7Z+1.
	*/

	if((*UserSecLen) % 6 != 1)
	{
		printf("WARNING: -> Enc_6By7MTRCode() -> For 6/7 MTR(2;8) code, sector length should be 6Z+1, where Z is an integer \n");
		*UserSecLen = *UserSecLen - ((*UserSecLen)%6) + 1;
		printf("WARNING: -> Enc_6By7MTRCode() -> Sector length changed to %d \n",(*UserSecLen));
	}
	UserSectorLength = *UserSecLen;

	NRZIUserBits = (int *) malloc((UserSectorLength-1) * sizeof(int));
	NRZIEncodedBits = (int *) malloc((int)((UserSectorLength-1)/rate) * sizeof(int));

	fCodewords = fopen("Codes/Rate6By7MTR2-8.dat","r");
	if(fCodewords == NULL)
	{
		printf("ERROR: -> Enc_6By7MTR2-8Code() -> Cannot open file %s \n","Codes/Rate6By7MTR2-8.dat");
		exit(-1);
	}
	
	/* Reading codewords from the file */
	for(i=0;i<64;i++)
	{
		fscanf(fCodewords,"%s\n",temp);
		for(j=0;j<7;j++)
		{
			dummy = temp[j];
			Codewords[i][j] = atoi(&(dummy));
		}
	}
	fclose(fCodewords);
	
	/* Convert NRZ userbits to NRZI userbits */
	for(i=1;i<UserSectorLength;i++)
	{
		NRZIUserBits[i-1]= abs(UserBits[i]-UserBits[i-1]);
	}

	/* Encoding the NRZI user bits */
	for(i=0;i<(UserSectorLength-1)/6;i++)
	{
		/* Decimal value of the 6-bit word */
		dec = 0;
		for(j=0;j<6;j++)
			dec = dec + (int) pow(2,5-j)*NRZIUserBits[i*6 + j];
		for(j=0;j<7;j++)
			NRZIEncodedBits[i*7 + j] = Codewords[dec][j];
	}

	/* Substitutions */
	for(i=0;i<(UserSectorLength-1)/6 - 1;i++)
	{
		/* Substitution Type I */
		if((NRZIEncodedBits[i*7+4]==0) && (NRZIEncodedBits[i*7+5]==0) && (NRZIEncodedBits[i*7+6]==1)
			&& (NRZIEncodedBits[(i+1)*7]==1) && (NRZIEncodedBits[(i+1)*7+1]==1) && (NRZIEncodedBits[(i+1)*7+2]==0))
		{
			NRZIEncodedBits[i*7+4] = 0;
			NRZIEncodedBits[i*7+5] = 1;
			NRZIEncodedBits[i*7+6] = 1;
			NRZIEncodedBits[(i+1)*7] = 0;
			NRZIEncodedBits[(i+1)*7 + 1] = 0;
			NRZIEncodedBits[(i+1)*7 + 2] = 1;
			continue;	/* Only one substitution will be required */
		}
		/* Substitution Type II */
		if((NRZIEncodedBits[i*7+4]==1) && (NRZIEncodedBits[i*7+5]==0) && (NRZIEncodedBits[i*7+6]==1)
			&& (NRZIEncodedBits[(i+1)*7]==1) && (NRZIEncodedBits[(i+1)*7+1]==1) && (NRZIEncodedBits[(i+1)*7+2]==0))
		{
			NRZIEncodedBits[i*7+4] = 0;
			NRZIEncodedBits[i*7+5] = 1;
			NRZIEncodedBits[i*7+6] = 1;
			NRZIEncodedBits[(i+1)*7] = 0;
			NRZIEncodedBits[(i+1)*7 + 1] = 1;
			NRZIEncodedBits[(i+1)*7 + 2] = 0;
			continue;	/* Only one substitution will be required */
		}
		/* Substitution Type III */
		if((NRZIEncodedBits[i*7+4]==0) && (NRZIEncodedBits[i*7+5]==0) && (NRZIEncodedBits[i*7+6]==0)
			&& (NRZIEncodedBits[(i+1)*7]==0) && (NRZIEncodedBits[(i+1)*7+1]==0) && (NRZIEncodedBits[(i+1)*7+2]==0))
		{
			NRZIEncodedBits[i*7+4] = 0;			
			NRZIEncodedBits[i*7+5] = 1;
			NRZIEncodedBits[i*7+6] = 1;
			NRZIEncodedBits[(i+1)*7] = 0;
			NRZIEncodedBits[(i+1)*7 + 1] = 0;
			NRZIEncodedBits[(i+1)*7 + 2] = 0;
			continue;	/* Only one substitution will be required */
		}
	}

	/* Convert NRZI Encoded Bits to NRZ format */
	/* Assume that the first NRZ encoded bit is 0 */
	EncodedBits[0] = 0;
	for(i=0;i<(int)((UserSectorLength-1)/rate);i++)
	{
		EncodedBits[i+1] = (NRZIEncodedBits[i] == 0) ? EncodedBits[i] : (EncodedBits[i]+1)%2;
	}
	free(NRZIUserBits);
	free(NRZIEncodedBits);
	return;
}


/*------------------------- ENCODER FOR 8/9 TMTR(2/3;11) CODE ------------------------------*/
/* 
This function implements the encoder for TMTR(2/3;11) rate 8/9 code.
For more info see: William Bliss, "An 8/9 rate time varying rellis code for
high density magnetic recording", IEEE Trans Mag, Vol 33, No 5, Sept 1997.
*/
void Enc_8By9TMTRCode(int *UserBits, int *EncodedBits, int *UserSecLen)
{
	int i,j,dec, UserSectorLength;
	char temp[9],dummy;
	int *NRZIUserBits,*NRZIEncodedBits;
	int Codewords[256][9];
	double rate = 8.0/9.0;
	FILE *fCodewords = NULL;
	
	/* 
	NRZ User Sector Length (*UserSecLen) should be of the form 8Z+1, where Z is an integer.
	Because, NRZI sector length would then be 8Z. NRZI Encoded length will be 9Z
	and NRZ encoded length will be 9Z+1.
	*/

	if((*UserSecLen) % 8 != 1)
	{
		printf("WARNING: -> Enc_8By9TMTRCode() -> For 8/9 TMTR(2/3;11) code, sector length should be 8Z+1, where Z is an integer \n");
		*UserSecLen = *UserSecLen - ((*UserSecLen)%8) + 1;
		printf("WARNING: -> Enc_8By9TMTRCode() -> Sector length changed to %d \n",(*UserSecLen));
	}
	UserSectorLength = *UserSecLen;

	NRZIUserBits = (int *) malloc((UserSectorLength-1) * sizeof(int));
	NRZIEncodedBits = (int *) malloc((int)((UserSectorLength-1)/rate) * sizeof(int));

	fCodewords = fopen("Codes/Rate8By9TMTR.dat","r");
	if(fCodewords == NULL)
	{
		printf("ERROR: -> Enc_8By9TMTRCode() -> Cannot open file %s \n","Codes/Rate8By9TMTR.dat");
		exit(-1);
	}
	
	/* Reading codewords from the file */
	for(i=0;i<256;i++)
	{
		fscanf(fCodewords,"%s\n",temp);
		for(j=0;j<9;j++)
		{
			dummy = temp[j];
			Codewords[i][j] = atoi(&(dummy));
		}
	}
	fclose(fCodewords);
	
	/* Convert NRZ userbits to NRZI userbits */
	for(i=1;i<UserSectorLength;i++)
	{
		NRZIUserBits[i-1]= abs(UserBits[i]-UserBits[i-1]);
	}

	/* Encoding the NRZI user bits */
	for(i=0;i<(UserSectorLength-1)/8;i++)
	{
		/* Decimal value of the 8-bit word */
		dec = 0;
		for(j=0;j<8;j++)
			dec = dec + (int) pow(2,7-j)*NRZIUserBits[i*8 + j];
		for(j=0;j<9;j++)
			NRZIEncodedBits[i*9 + j] = Codewords[dec][j];
	}

	/* Convert NRZI Encoded Bits to NRZ format */
	/* Assume that the first NRZ encoded bit is 0 */
	EncodedBits[0] = 0;
	for(i=0;i<(int)((UserSectorLength-1)/rate);i++)
	{
		EncodedBits[i+1] = (NRZIEncodedBits[i] == 0) ? EncodedBits[i] : (EncodedBits[i]+1)%2;
	}

	free(NRZIUserBits);
	free(NRZIEncodedBits);

	return;
}

/*------------------------- COUNT TRIBITS ----------------------------------*/
/* Calculates the number of possible +-+ and -+- error events in the vector */
int CountPossibleErrorEvents(int *InputWord, int Length)
{
	int i,NConsTrans,NumErrorEvents=0;
	for(i=1;i<Length;i++)
	{
		NConsTrans=0;
		while((abs(InputWord[i]-InputWord[i-1])==1)&&(i<Length))
		{
			NConsTrans++;
			i++;
		}
		if(NConsTrans>=3)
			NumErrorEvents = NumErrorEvents + (NConsTrans-1);
	}
	return(NumErrorEvents);
}


/*------------------------- PERMUTATION ENOCDER CODE ------------------------------*/
/* Given a permutation with period p, and an input codeword, 
this function permutes the codeword p possible times and chooses the one with lowest number
of tribits. Returns the number of times the permutation was applied and the permuted codeword.
*/
#define DEBUG_PERMUTE 0

int Permute(int *InputWord, int *PermutedWord, int SectorLength, int *P, int PLength, int Period)
{
	int MinNumErrorEvents=0;	/* Some big number */
	int NumPermutations=0;
	int i,j,NumErrorEvents;

	MinNumErrorEvents = CountPossibleErrorEvents(InputWord, SectorLength);

	if(DEBUG_PERMUTE)
	{
		printf("\nPermutation: [ ");
		for(j=0;j<PLength;j++)
			printf("%d ",P[j]);
		printf("]\n\n");

		for(j=0;j<SectorLength;j++)
			printf("%d ",InputWord[j]);
		printf("\n");
		printf("%d\n\n", MinNumErrorEvents);
	}
	for(j=0;j<SectorLength;j++)
		PermutedWord[j] = InputWord[j];
	
	/* Permutation P is in cyclic format, i.e. InputWord(P(1)) becomes InputWord(P(2)),
	   InputWord(P(2)) becomes InputWord(P(3)) etc.
	   Applying P 'Period' times = Identity transformation
	*/
	for(i=1;i<=Period-1;i++)
	{
		/* Apply permutation P^i to the InputWord */
		for(j=0;j<=PLength-1;j++)
			PermutedWord[P[(j+i)%(PLength)]] = InputWord[P[j]];

		NumErrorEvents = CountPossibleErrorEvents(PermutedWord, SectorLength);
		if(NumErrorEvents < MinNumErrorEvents)
		{
			MinNumErrorEvents = NumErrorEvents;
			NumPermutations = i;
		}
		if(DEBUG_PERMUTE)
		{
			for(j=0;j<SectorLength;j++)
				printf("%d ",PermutedWord[j]);
			printf("\n");
			printf("%d\n\n",NumErrorEvents);
		}		
	}
	
	/* Encoded word */
	for(j=0;j<=PLength-1;j++)
		PermutedWord[P[(j+NumPermutations)%(PLength)]] = InputWord[P[j]];	

	if(DEBUG_PERMUTE)
	{
		printf("Number of Permutation applied = %d\n",NumPermutations);
		printf("Number of error events possible in the encoded word = %d\n", MinNumErrorEvents);
	}
	return(NumPermutations);
}


